# -*- coding: utf-8 -*-

#Created on Thu Feb 01 08:56:54 2018

#@author: ryanday
#MIT License

#Copyright (c) 2018 Ryan Patrick Day

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A ▲ICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

import numpy as np
import matplotlib.pyplot as plt
from matplotlib import rcParams
import matplotlib.cm as cm
import chinook.klib as K_lib
import chinook.Ylm as Ylm

'''
Library for different operators of possible interest in calculating, diagnostics, etc for a material of interest

'''

def LSmat(TB,axis=None):
    '''
    Generate an arbitary L.S type matrix for a given basis. Uses same *Yproj* as 
    the *HSO* in the *chinook.H_library*, but is otherwise different, as it supports
    projection onto specific axes, in addition to the full vector dot product operator. 

    Otherwise, the LiSi matrix is computed with i the axis index. 
    To do this, a linear combination of L+S+,L-S-,L+S-,L-S+,LzSz terms are used to compute.
    
    In the factors dictionary, the weight of these terms is defined. 
    The keys are tuples of (L+/-/z,S+/-/z) in a bit
    of a cryptic way. For L, range (0,1,2) ->(-1,0,1) 
    and for S range (-1,0,1) = S1-S2 with S1/2 = +/- 1 here
    
    L+,L-,Lz matrices are defined for each l shell in the basis, 
    transformed into the basis of cubic harmonics.
    The nonzero terms will then just be used along with the spin and 
    weighted by the factor value, and slotted into 
    a len(basis)xlen(basis) matrix HSO
    
    *args*:

        - **TB**: tight-binding object, as defined in TB_lib.py
        
        - **axis**: axis for calculation as either 'x','y','z',None,
        or float (angle in the x-y plane)
        
    *return*:

        - **HSO**: (len(basis)xlen(basis)) numpy array of complex float
        
    ***
    '''
    
    Md = Ylm.Yproj(TB.basis)
    normal_order = {0:{'':0},1:{'x':0,'y':1,'z':2},2:{'xz':0,'yz':1,'xy':2,'ZR':3,'XY':4},3:{'z3':0,'xz2':1,'yz2':2,'xzy':3,'zXY':4,'xXY':5,'yXY':6}}
    factors = {(2,-1):0.5,(0,1):0.5,(1,0):1.0}
    L,al = {},[]
    HSO = np.zeros((len(TB.basis),len(TB.basis)),dtype=complex)
    for o in TB.basis:
        if (o.atom,o.n,o.l) not in al:
            al.append((o.atom,o.n,o.l))
            Mdn = Md[(o.atom,o.n,o.l,-1)]
            Mup = Md[(o.atom,o.n,o.l,1)]
            Mdnp = np.linalg.inv(Mdn)
            Mupp = np.linalg.inv(Mup)
            L[(o.atom,o.n,o.l)] = [np.dot(Mupp,np.dot(Lm(o.l),Mdn)),np.dot(Mdnp,np.dot(Lz(o.l),Mdn)),np.dot(Mdnp,np.dot(Lp(o.l),Mup))]
    if axis is not None:
        try:
            ax = float(axis)
            factors = {(0,1):0.25,(2,-1):0.25,(2,1):0.25*np.exp(-1.0j*2*ax),(0,-1):0.25*np.exp(1.0j*2*ax)}
        except ValueError:
            if axis=='x':
                factors = {(0,1):0.25,(2,-1):0.25,(2,1):0.25,(0,-1):0.25}
            elif axis=='y':
                factors = {(0,1):0.25,(2,-1):0.25,(2,1):-0.25,(0,-1):-0.25}
            elif axis=='z':
                factors = {(1,0):1.0}
            else:
                print('Invalid axis entry')
                return None
    for o1 in TB.basis:
        for o2 in TB.basis:
            if o1.index<=o2.index:
                LS_val = 0.0
                if np.linalg.norm(o1.pos-o2.pos)<0.0001 and o1.l==o2.l and o1.n==o2.n:
                    inds = (normal_order[o1.l][o1.label[2:]],normal_order[o2.l][o2.label[2:]])
                    
                    ds = (o1.spin-o2.spin)/2.
                    if ds==0:
                        s=0.5*np.sign(o1.spin)
                    else:
                        s=1.0
                    for f in factors:
                        if f[1]==ds:
                            LS_val+=factors[f]*L[(o1.atom,o1.n,o1.l)][f[0]][inds]*s
                    HSO[o1.index,o2.index]+=LS_val
                    if o1.index!=o2.index:
                        HSO[o2.index,o1.index]+=np.conj(LS_val)
    return HSO

        

def Lp(l):
    
    '''
    
    L+ operator in the l,m_l basis, organized with 
    (0,0) = |l,l>, (2*l,2*l) = |l,-l>
    The nonzero elements are on the upper diagonal
    
    *arg*: 

        - **l**: int orbital angular momentum
        
    *return*:

        - **M**: numpy array (2*l+1,2*l+1) of real float
    ***
    '''
    
    M = np.zeros((2*l+1,2*l+1))
    r = np.arange(0,2*l,1)
    M[r,r+1]=1.0
    vals = [0]+[np.sqrt(l*(l+1)-(l-m)*(l-m+1)) for m in range(1,2*l+1)]
    M = M*vals
    return M

def Lm(l):
    
    '''
    
    L- operator in the l,m_l basis, organized with 
    (0,0) = |l,l>, (2*l,2*l) = |l,-l>
    The nonzero elements are on the upper diagonal
    
    *arg*:

        - **l**: int orbital angular momentum
    
    *return*:

        - **M**: numpy array (2*l+1,2*l+1) of real float
        
    ***
    '''
    
    M = np.zeros((2*l+1,2*l+1))
    r = np.arange(1,2*l+1,1)
    M[r,r-1]=1.0
    vals = [np.sqrt(l*(l+1)-(l-m)*(l-m-1)) for m in range(0,2*l)]+[0]
    M = M*vals
    return M

def Lz(l):
    
    '''
    
    Lz operator in the l,m_l basis
    
    *arg*:

        - **l**: int orbital angular momentum
        
    *return*:

        - numpy array (2*l+1,2*l+1)
        
    ***
    '''
    return np.identity(2*l+1)*np.array([l-m for m in range(2*l+1)])


def fatbs(proj,TB,Kobj=None,vlims=(0,1),Elims=(-1,1),degen=False):
    
    '''
    
    Fat band projections. User denotes which orbital index projection is of interest
    Projection passed either as an Nx1 or Nx2 array of float. If Nx2, first column is
    the indices of the desired orbitals, the second column is the weight. If Nx1, then
    the weights are all taken to be eqaul
    
    *args*:

        - **proj**: iterable of projections, to be passed as either a 1-dimensional
        with indices of projection only, OR, 2-dimensional, with the second column giving
        the amplitude of projection (for linear-combination projection)
            
        - **TB**: tight-binding object
            
    *kwargs*:

        - **Kobj**: Momentum object, as defined in *chinook.klib.py*
            
        - **vlims**: tuple of 2 float, limits of the colorscale for plotting, default to (0,1)
            
        - **Elims**: tuple of 2 float, limits of vertical scale for plotting
            
        - **degen**: bool, True if bands are degenerate, sum over adjacent bands
            
    *return*:

        - **Ovals**: numpy array of float, len(Kobj.kpts)*len(TB.basis)
        
    ***
    '''
    
    proj = np.array(proj)
    
    O = np.identity(len(TB.basis))
    
    if len(np.shape(proj))<2:
        tmp = np.zeros((len(proj),2))
        tmp[:,0] = proj
        tmp[:,1] = 1.0
        proj = tmp
    pvec = np.zeros(len(TB.basis),dtype=complex)
    try:
        pvec[np.real(proj[:,0]).astype(int)] = proj[:,1]
        O = O*pvec
    
        Ovals = O_path(O,TB,Kobj,vlims,Elims,degen=degen)
    except ValueError:
        print('projections need to be passed as list or array of type [index,projection]')
    
        Ovals = None
    
    return Ovals
    
    


def O_path(Operator,TB,Kobj=None,vlims=(0,0),Elims=(-10,10),degen=False,plot=True,ax=None):
    
    '''
    
    Compute and plot the expectation value of an user-defined operator along a k-path
    Option of summing over degenerate bands (for e.g. fat bands) with degen boolean flag
        
    *args*:

        - **Operator**: matrix representation of the operator (numpy array len(basis), len(basis) of complex float)
          
        - **TB**: Tight binding object from TB_lib
            
    *kwargs*:
        
        - **Kobj**: Momentum object, as defined in *chinook.klib.py*
            
        - **vlims**: tuple of 2 float, limits of the colourscale for plotting, default to (0,0)
        if default value passed, will compute a reasonable range
            
        - **Elims**: tuple of 2 float, limits of vertical scale for plotting
            
        - **degen**: bool, True if bands are degenerate, sum over adjacent bands
        
        - **ax**: matplotlib Axes, option for plotting onto existing Axes

    *return*:

        - **O_vals**: the numpy array of float, (len Kobj x len basis) expectation values 
        
        - **ax**: matplotlib Axes, allowing for user to further modify
    
    ***
    '''
    
    if np.shape(Operator)!=(len(TB.basis),len(TB.basis)):
        print('ERROR! Ensure your operator has the same dimension as the basis.')
        return None
    try:
        np.shape(TB.Evec)
    except AttributeError:
        try:
            len(TB.Kobj.kpts)
            TB.solve_H()
        except AttributeError:    
            TB.Kobj = Kobj
            try:
                TB.solve_H()
            except TypeError:
                print('ERROR! Please include a K-object, or diagonalize your tight-binding model over a k-path first to initialize the eigenvectors')
                return None
            
    #check if Hermitian
    
       
    right_product = np.einsum('ij,ljm->lim',Operator,TB.Evec)
    O_vals = np.einsum('ijk,ijk->ik',np.conj(TB.Evec),right_product)
    O_vals = np.real(O_vals) #any Hermitian operator must have real-valued expectation value--discard any imaginary component
    if degen:
        O_vals = degen_Ovals(O_vals,TB.Eband)


    rcParams.update({'font.size':14})
    
    if ax is None:
        fig = plt.figure()
        fig.set_tight_layout(False)
        ax=fig.add_subplot(111)

    for b in TB.Kobj.kcut_brk:
        plt.axvline(x = b,color = 'grey',ls='--',lw=1.0)
                    
    if np.sign(O_vals.min())<0 and np.sign(O_vals.max())>0:
        my_cmap = cm.RdBu
    else:
        my_cmap = cm.Blues
    
    if vlims==(0,0):
        vlims = (O_vals.min()-(O_vals.max()-O_vals.min())/10.0,O_vals.max()+(O_vals.max()-O_vals.min())/10.0)
    if Elims==(-10,10):
        Elims = (TB.Eband.min()-(TB.Eband.max()-TB.Eband.min())/10.0,TB.Eband.max()+(TB.Eband.max()-TB.Eband.min())/10.0)
    if plot:
        for p in range(np.shape(O_vals)[1]):

            plt.plot(TB.Kobj.kcut,TB.Eband[:,p],color='k',lw=0.2)
            O_line=plt.scatter(TB.Kobj.kcut,TB.Eband[:,p],c=O_vals[:,p],cmap=my_cmap,marker='.',lw=0,s=80,vmin=vlims[0],vmax=vlims[1])

        plt.axis([TB.Kobj.kcut[0],TB.Kobj.kcut[-1],Elims[0],Elims[1]])
        plt.xticks(TB.Kobj.kcut_brk,TB.Kobj.labels)
        plt.colorbar(O_line,ax=ax)
        plt.ylabel("Energy (eV)")
        
    return O_vals,ax


def degen_Ovals(Oper_exp,Energy):
    
    '''
    In the presence of degeneracy, we want to average over the
    evaluated orbital expectation values--numerically, the degenerate 
    subspace can be arbitrarily diagonalized during numpy.linalg.eigh. 
    All degeneracies are found, and the expectation values averaged.
    
    *args*:

        - **Oper_exp**: numpy array of float, operator expectations
        
        - **Energy**: numpy array of float, energy eigenvalues.
    
    ***
    '''
    
    O_copy = Oper_exp.copy()
    tol = 1e-9
    for ki in range(np.shape(Oper_exp)[0]):
        val = Energy[ki,0]
        start = 0
        counter = 1
        for bi in range(1,np.shape(Oper_exp)[1]):
            if abs(Energy[ki,bi]-val)<tol:
                counter+=1

            if abs(Energy[ki,bi]-val)>=tol or bi==(np.shape(Oper_exp)[1]-1):

                O_copy[ki,start:start+counter] = np.mean(O_copy[ki,start:start+counter])
                start = bi
                counter = 1
                val = Energy[ki,bi]
    return O_copy                

def O_surf(O,TB,ktuple,Ef,tol,vlims=(0,0),ax=None):
    
    '''
    
    Compute and plot the expectation value of an user-defined operator over
    a surface of constant-binding energy
    
    Option of summing over degenerate bands (for e.g. fat bands) with degen boolean flag
        
    *args*:

        - **O**: matrix representation of the operator (numpy array len(basis), len(basis) of complex float)
            
        - **TB**: Tight binding object from *chinook.TB_lib.py*
        
        - **ktuple**: momentum range for mesh: 
            ktuple[0] = (x0,xn,n),ktuple[1]=(y0,yn,n),ktuple[2]=kz
            
    *kwargs*:

        - **vlims**: limits for the colourscale (optional argument), will choose 
        a reasonable set of limits if none passed by user
        
        - **ax**: matplotlib Axes, option for plotting onto existing Axes


    *return*:

        - **pts**: the numpy array of expectation values, of shape Nx3, with first
        two dimensions the kx,ky coordinates of the point, and the third the expectation
        value.
        
        - **ax**: matplotlib Axes, allowing for further user modifications
        
    ***
    '''

    
    coords,Eb,Ev=FS(TB,ktuple,Ef,tol)
    masked_Ev = np.array([Ev[int(coords[ei,2]/len(TB.basis)),:,int(coords[ei,2]%len(TB.basis))] for ei in range(len(coords))])
    Ovals = np.sum(np.conj(masked_Ev)*np.dot(O,masked_Ev.T).T,1)

    if np.sign(Ovals.min())!=np.sign(Ovals.max()):
        cmap = cm.RdBu
    else:
        cmap = cm.magma
    
    pts = np.array([[coords[ci,0],coords[ci,1],np.real(Ovals[ci])] for ci in range(len(coords))])
    
    if vlims==(0,0):
        vlims = (Ovals.min()-(Ovals.max()-Ovals.min())/10.0,Ovals.max()+(Ovals.max()-Ovals.min())/10.0)

    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.scatter(pts[:,0],pts[:,1],c=pts[:,2],cmap=cmap,s=200,vmin=vlims[0],vmax=vlims[1])
    ax.scatter(pts[:,0],pts[:,1],c='k',s=5)
    
    return pts,ax



def FS(TB,ktuple,Ef,tol,ax=None):
    
    '''
    A simplified form of Fermi surface extraction, for proper calculation of this,
    *chinook.FS_tetra.py* is preferred. This finds all points in kmesh within a 
    tolerance of the constant energy level.
    
    *args*:

        - **TB**: tight-binding model object
        
        - **ktuple**: tuple of k limits, len (3), First and second should be iterable,
        define the limits and mesh of k for kx,ky, the third is constant, float for kz
        
        - **Ef**: float, energy of interest, eV
        
        - **tol**: float, energy tolerance, float
        
        - **ax**: matplotlib Axes, option for plotting onto existing Axes

    
    *return*:

        - **pts**: numpy array of len(N) x 3 indicating x,y, band index
        
        - **TB.Eband**: numpy array of float, energy spectrum
        
        - **TB.Evec**: numpy array of complex float, eigenvectors
        
        - **ax**: matplotlib Axes, for further user modification
    
    ***
    '''
    
    x,y,z=np.linspace(*ktuple[0]),np.linspace(*ktuple[1]),ktuple[2]
    X,Y=np.meshgrid(x,y)
        
    k_arr,_ = K_lib.kmesh(0.0,X,Y,z)  

    blen = len(TB.basis)    

    TB.Kobj = K_lib.kpath(k_arr)
    _,_ = TB.solve_H()
    TB.Eband = np.reshape(TB.Eband,(np.shape(TB.Eband)[-1]*np.shape(X)[0]*np.shape(X)[1])) 
    pts = []
    for ei in range(len(TB.Eband)):
        if abs(TB.Eband[ei]-Ef)<tol: 
            inds = (int(np.floor(np.floor(ei/blen)/np.shape(X)[1])),int(np.floor(ei/blen)%np.shape(X)[1]))
            pts.append([X[inds],Y[inds],ei])
    
    pts = np.array(pts)
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.scatter(pts[:,0],pts[:,1])
    return pts,TB.Eband,TB.Evec,ax
    
    
####################SOME STANDARD OPERATORS FOLLOW HERE: ######################
    
    

def LdotS(TB,axis=None):
    '''
    Wrapper for **O_path** for computing L.S along a vector projection of interest,
    or none at all.
    
    
    *args*:

        - **TB**: tight-binding obect
        
    *kwargs*:

        - **axis**: numpy array of 3 float, indicating axis, or None for full L.S
    
    *return*:

        - **O**: numpy array of Nxlen(basis) float, expectation value of operator
        on each band over the kpath of TB.Kobj.
    
    ***
    '''
    HSO = LSmat(TB,axis)
    O = O_path(HSO,TB,TB.Kobj)
    return O

def Sz(TB):
    '''
    Wrapper for **O_path** for computing Sz along a vector projection of interest,
    or none at all.
    
    
    *args*:

        - **TB**: tight-binding obect
        
    
    *return*:

        - **O**: numpy array of Nxlen(basis) float, expectation value of operator
        on each band over the kpath of TB.Kobj.
    
    ***
    '''
    Omat = S_vec(len(TB.basis),np.array([0,0,1]))
    O = O_path(Omat,TB,TB.Kobj)
    return O



def surface_proj(basis,length):
    '''
    Operator for computing surface-projection of eigenstates. User passes the orbital basis
    and an extinction length (1/e) length for the 'projection onto surface'. The operator 
    is diagonal with exponenential suppression based on depth.
    
    For use with SLAB geometry only
    
    *args*:

        - **basis**: list, orbital objects
        
        - **cutoff**: float, cutoff length
    
    *return*:

        - **M**: numpy array of float, shape len(TB.basis) x len(TB.basis)
        
    ***
    '''
    Omat = np.identity(len(basis))
    attenuation = np.exp(np.array([-abs(o.depth)/length for o in basis]))
    return Omat*attenuation
    

def S_vec(LB,vec):
    '''
    Spin operator along an arbitrary direction can be written as
    n.S = nx Sx + ny Sy + nz Sz
    
    *args*:

        - **LB**: int, length of basis
        
        - **vec**: numpy array of 3 float, direction of spin projection
    
    *return*:

        - numpy array of complex float (LB by LB), spin operator matrix
    
    ***
    '''
    numstates = int(LB/2)
    Si = 0.5*np.identity(numstates,dtype=complex)
    
    Smats = np.zeros((3,LB,LB),dtype=complex)
    Smats[2,:numstates,:numstates]+=-Si
    Smats[2,numstates:,numstates:]+=Si
    Smats[0,:numstates,numstates:] +=Si
    Smats[0,numstates:,:numstates]+=Si
    Smats[1,:numstates,numstates:]+=1.0j*Si
    Smats[1,numstates:,:numstates]+=-1.0j*Si
    
    return vec[0]*Smats[0]+vec[1]*Smats[1]+vec[2]*Smats[2]

    


def is_numeric(a):

    '''
    Quick check if object is numeric
    
    *args*:

        - **a**: numeric, float/int
    
    *return*:

        - bool, if numeric True, else False

    ***
    '''
    if a is not None:
        
        try:
            float(a)
            return True
        except ValueError:
            return False
    else:
        return False
    
    


# -*- coding: utf-8 -*-

#Created on Sun Jul 15 13:49:17 2018


#@author: rday

#MIT License

#Copyright (c) 2018 Ryan Patrick Day

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.
#The following algorithm is based very closely off of the work of 
#W. Sun and G. Ceder, Surface Science 617, 53 (2013)

import numpy as np

def ang_v1v2(v1,v2):
    '''
    Find angle between two vectors, rounded to floating point precision.

    *args*:

        - **v1**: numpy array of N float

        - **v2**: numpy array of N float

    *return*: 

        - float, angle in radians

    ***
    '''
    return np.arccos(np.around(np.dot(v1,v2)/(np.linalg.norm(v1)*np.linalg.norm(v2)),15))

def are_parallel(v1,v2):
    '''
    Are two vectors parallel?

    *args*:

        - **v1**: numpy array of N float

        - **v2**: numpy array of N float

    *return*:

        - boolean, True if parallel, to within 1e-5 radians, False otherwise

    ***
    '''
    cos = np.dot(v1,v2)/(np.linalg.norm(v1)*np.linalg.norm(v2))
    if abs(cos)>1e-5 and np.mod(cos,1)<1e-5:
        return True
    else:
        return False

def are_same(v1,v2):
  
    '''
    Are two vectors identical, i.e. parallel and of same length, to within 
    the precision of *are_parallel*?

    *args*:

        - **v1**: numpy array of N float

        - **v2**: numpy array of N float

    *return*:

        - boolean, True if identical, False otherwise.

    ***
    '''
    if are_parallel(v1,v2):
        mod_val = (np.mod(np.dot(v1,v2)/np.linalg.norm(v2)**2,1))
        if mod_val<1.0e-5 or mod_val>0.99999:
            return True
        else:
            return False
    else:
        return False
    
    
def initialize_search(v1,v2,avec):
    '''
    Seed search for v3 with the nearest-neighbouring Bravais latice point which maximizes
    the projection out of plane of that spanned by v1 and v2

    *args*:

        - **v1**: numpy array of 3 float, a spanning vector of the plane

        - **v2**: numpy array of 3 float, a spanning vector of the plane
    
        - **avec**: numpy array of 3x3 float

    *return*:

        - numpy array of float, the nearby Bravais lattice point which maximizes
         the projection along the plane normal

    ***
    '''
    nv = np.cross(v1,v2)
    near = np.array([np.dot([int(np.mod(i,(3)))-1,int(np.mod(i/(3),(3)))-1,int(i/(3)**2)-1],avec) for i in range((3)**3) if i!=int((3)**3/2)])
    angs = np.array([ang_v1v2(ni,nv) for ni in near])
    choice = np.where(abs(angs)==abs(angs).min())[0]
    return near[choice[0]]


def refine_search(v3i,v1,v2,avec,maxlen):
    '''
    Refine the search for the optimal v3 which both minimalizes the length while
    maximizing orthogonality to v1 and v2

    *args*:

        - **v3i**: numpy array of 3 float, initial guess for the surface vector

        - **v1**: numpy array of 3 float, a spanning vector of the plane

        - **v2**: numpy array of 3 float, a spanning vector of the plane

        - **avec**: numpy array of 3x3 float

        - **maxlen**: float, longest vector accepted

    *return*:

        **v3_opt**: list of numpy array of 3 float, options for surface vector
    
    ***
    '''
    nv = np.cross(v1,v2)
    atol = 1e-2
    v_add = np.array([-v2,-v1,np.zeros(3),v1,v2])
    v3_opt = []
    for qi in range(1,50):
        tmp_v = qi*v3i
        ang_to_norm = ang_v1v2(tmp_v,nv)
        if abs(ang_to_norm)<atol:
            ok = True
        else:
            ok = False
        counter = 1
        while not ok:
            v_opt = tmp_v + counter*v_add
            angles = np.array([ang_v1v2(vi,nv) for vi in v_opt])
            try:
                better = abs(angles[np.where(abs(angles)<abs(ang_to_norm))[0]]).min()
                tmp_v = v_opt[np.where(abs(angles)==better)[0][0]]
                ang_to_norm = ang_v1v2(tmp_v,nv)
                if abs(ang_v1v2(tmp_v,nv))<atol:
                    ok = True

            except ValueError:
                ok = True
            counter+=1
        if np.linalg.norm(tmp_v)<maxlen:
            v3_opt.append(tmp_v)
    return v3_opt

def score(vlist,v1,v2,avec):

    '''

    The possible surface vectors are scored based on their legnth and their orthogonality 
    to the in-plane vectors.

    *args*:

        - **vlist**: list fo numpy array of 3 float, options for surface vector

        - **v1**: numpy array of 3 float, a spanning vector of the plane

        - **v2**: numpy array of 3 float, a spanning vector of the plane

        - **avec**: numpy array of 3x3 float

    *return*:

        - numpy array of 3 float, the best scoring vector option

    ***
    '''
    nv = np.cross(v1,v2)
    
    angles = np.array([ang_v1v2(vi,nv) for vi in vlist])
    if abs(angles).max()>0.0:
        angles/=abs(angles).max()
    len_proj = np.array([np.linalg.norm(vi) for vi in vlist])
    if abs(len_proj).max()>0.0:
        len_proj/=abs(len_proj).max()
    score_vec = np.array([[len_proj[i],angles[i]] for i in range(len(vlist))])
    return vlist[np.where(np.linalg.norm(score_vec)==np.linalg.norm(score_vec).min())[0][0]]
                    

def find_v3(v1,v2,avec,maxlen):
    '''
    Wrapper function for finding the surface vector. 

    *args*:

        - **v1**: numpy array of 3 float, a spanning vector of the plane

        - **v2**: numpy array of 3 float, a spanning vector of the plane

        - **avec**: numpy array of 3x3 float

        - **maxlen**: float, longest accepted surface vector
        
    *return*: 

        - numpy array of 3 float, surface vector choice  

    ***  
    '''
    v3i = initialize_search(v1,v2,avec)
    v3f = refine_search(v3i,v1,v2,avec,maxlen)
    v3_choice = score(v3f,v1,v2,avec)
    return v3_choice
                    
            
    


# -*- coding: utf-8 -*-

#Created on Wed Sep 28 10:15:56 2016

#@author: ryanday
#MIT License

#Copyright (c) 2018 Ryan Patrick Day

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.



#Slater's rules for calculating effective nuclear charge:
#   Group --- Other electrons in --- electrons in group(s) w/ --- electrons in group(s) --- electrons in group(s) w/
#               the same group            n'=n, l'<l                   w/ n'=n-1                    n'<=n-2
# -------------------------------------------------------------------------------------------------------------------
#  [1s]    ---        0.30        ---         -                ---          -            ---            -
#  [ns,np] ---        0.35        ---         -                ---         0.85          ---            1
#  [nd,nf] ---        0.35        ---         1                ---          1            ---            1

#Effective nuclear charge: n = 1,2,3,4,5,6 --- n' = 1,2,3,3.7,4.0,4.2


import numpy as np
import scipy.special as sc
import matplotlib.pyplot as plt
import pkg_resources
import linecache
from math import factorial
from math import gamma

import chinook.Ylm as Y
import chinook.atomic_mass as am

ao = 5.29177*10**-1 #in units of angstroms
me = 9.11*10**-31
hb = 6.626e-34/2/np.pi
q = 1.602e-19
Eo = 8.85e-12
mp = 1.67*10**-27
A = 10**-10

ndic={1:1,2:2,3:3,4:3.7,5:4,6:4.2}


textnm="electron_configs.txt"

filename = pkg_resources.resource_filename(__name__,textnm)

def get_con(filename,Z):
    
    '''
    Get electron configuration for a given element, from electron_configs.txt.
    This is configuration to be used in calculation of the Slater wavefunction.
    
    *args*:

        - **filename**: string, text-file where the configurations are saved
        
        - **Z**: int, atomic number
        
    *return*:

        - string, electron configuration of neutral atom
    
    ***
    '''
    try:
        return linecache.getline(filename,int(Z)).split(',')[1].strip()
    except IndexError:
        print('ERROR: Invalid atomic number, returning  nothing')
        return ''

      
def shield_split(shield_string):
    
    '''

    Parse the electron configuration string, dividing up into the different orbital
    shells.
    
    *args*:

        - **shield_string**: string, electron configuration
        
    *return*:

        - **parsed_shield**: list of separated orbital components
    
    ***
    '''
    alphas = []
    for s in range(len(shield_string)):
        if shield_string[s].isalpha():
            alphas.append(s)
    parsed_shield = []
    for s in range(len(alphas)-1):
        parsed_shield.append(shield_string[alphas[s]-1:alphas[s+1]-1])
    parsed_shield.append(shield_string[alphas[-1]-1:])
    return parsed_shield
        
def Z_eff(Z_ind,orb):
    
    '''
    Compute the effective nuclear charge, following Slater's rules.
    
    *args*:

        - **Z_ind**: int, the atomic number
        
        - **orb**: orbital string, written as nl in either fully numeric, or numeric-alpha format
    
    *return*:

        - **result**: float, effective nuclear charge
    
    ***
    '''
    
    l_dic = {"s":0,"p":1,"d":2,"f":3}  
    l_dic_inv = {0:"s",1:"p",2:"d",3:"f"}
    e_conf = get_con(filename,Z_ind)
    n = int(orb[0])
    if orb[1].isalpha():
        l=orb[1] # if orbital string already written in form of 2p for example, rather than 21, simply take l as 'p'
    elif not orb[1].isalpha():#if its not written as 2p, but rather 21, then convert the 1 into p and then rewrite orb as '2p'
        l=int(orb[1])
        tmp = orb[0]+l_dic_inv[l]
        orb = tmp
    for s in range(len(e_conf)-1): #iterate over the character string for the electronic configuration from table
        if e_conf[s]+e_conf[s+1]==orb: #if we find the orbital string, stop
            shield=e_conf[:s] #take the shield as all preceding elements of the configuration
            val = e_conf[s:s+3] # the valence are the remaining
            if orb[-1]=="s": #for the case of an s valence, then you need to add the next set of stuff too
                shield+=e_conf[s+3:s+6]
    try:
        parsed_shield=shield_split(shield)
    except UnboundLocalError:
        print('ERROR: Invalid orbital combination given for Z = {:d}, returning None'.format(Z_ind))
        return None
    nval = int(val[0])
    lval = l_dic[val[1]]
    fill_val = int(val[2:])
    s_sum = 0.0
    if nval==1:
        s_sum+=0.3*(fill_val-1)
    if nval>1:
        for i in range(len(parsed_shield)):
            tmp = parsed_shield[i]
            n = int(tmp[0])
            l = l_dic[tmp[1]]
            fill = int(tmp[2:])
            if n<=nval-2:
                s_sum+=1.0*fill
            elif n==nval-1:
                if lval<=1:
                    s_sum+=0.85*fill
                elif lval>1:
                    s_sum+=1.0*fill
            if n==nval and l<lval and lval>1:
                s_sum+=1.0*fill
            elif n==nval and lval<2:
                s_sum+=0.35*fill
        s_sum+=0.35*(fill_val-1) #using Slater's rules from above to compute the effective screening
    result = Z_ind-s_sum
    return result
    

def Slater_exec(Z_ind,orb):

    '''

    Define an executable Slater-type orbital wavefunction which takes only
    the radius as an input argument. In this way, the usser specifies Z 
    and the orbital label string, and this generates a lambda function
    
    :math:`R(r) = (\\frac{2Z_{e}}{n_{e}})^{n_{e}} \\sqrt{\\frac{2Z_{e}}{n_{e}\\Gamma(2n_{e}+1)}} r^{n_{e}-1} e^{-\\frac{Z_{e} r}{n_{e}}}`    
 
    *args*:

        - **Z_ind**: int, atomic number
        
        - **orb**: string, 'nlxx' as per standard orbital definitions used
        throughout the library
    
    *return*:

        - executable function of position (in units of Angstrom)
    
    ***
    '''
    ne = ndic[int(orb[0])]
    xi = Z_eff(Z_ind,orb)/ne
    tmp = (2*xi)**ne*np.sqrt(2*xi/gamma(float(2*ne)+1))
    
    def lambda_gen():

        return lambda r: tmp*(r)**(ne-1)*np.exp(-xi*r)
    return lambda_gen()
  
    


def hydrogenic_exec(Z_ind,orb):
    
    '''
    
    Similar to Slater_exec, we define an executable function for the 
    hydrogenic orbital related to the orbital defined with atomic number
    **Z_ind** and orbital label 'nlxx'
    
    *args*:

        - **Z_ind**: int, atomic number
        
        - **orb**: string, orbital label 'nlxx'
        
    *return*:
    
        - executable function of float
        
    ***
    '''
    
    mn = am.get_mass_from_number(Z_ind)
    au = (4*np.pi*Eo*hb**2*(mn*mp+me))/(q**2*mn*mp*me)
    n = int(orb[0])
    l = int(orb[1])
    
    def lambda_gen():
        return lambda r: (np.sqrt((2*Z_ind*A/(n*au))**3*factorial(n-l-1)/(2*n*factorial(n+l)))*np.exp(-Z_ind*r*A/(n*au))*(2*Z_ind*r*A/(n*au))**l*sc.genlaguerre(n-l-1,l*2+1)(2*Z_ind*r*A/(n*au)))
    
    return lambda_gen()

if __name__=="__main__":
    mn = am.get_mass_from_number(6)
    r = np.linspace(0,10,1000)
    Chyd = hydrogenic_exec(6,'21z')(r)
    CSlat= Slater_exec(6,'21x')(r)
    plt.figure()
    plt.plot(r,Chyd)
    plt.plot(r,CSlat)
    



# -*- coding: utf-8 -*-

#Created on Fri Apr 27 12:37:56 2018
#@author: ryanday
#MIT License

#Copyright (c) 2018 Ryan Patrick Day

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.



import numpy as np

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.cm as cm
import matplotlib.tri as mtri

import chinook.Ylm as Ylm






class wavefunction:
    '''
    This class acts to reorganize basis and wavefunction information in a more
    suitable data structure than the native orbital class, or the sake of plotting
    orbital wavefunctions. The relevant eigenvector can be redefined, so long as it
    represents a projection onto the same orbital basis set as defined previously.
    
    *args*:

        - **basis**: list of orbital objects
        
        - **vector**: numpy array of complex float, eigenvector projected onto the basis orbitals
   
    '''
    
    def __init__(self,basis,vector):
        if len(basis)==len(vector):
            self.basis = basis
            self.centres,self.centre_pointer = self.find_centres()
            self.harmonics,self.harmonic_pointer,self.projections =self.find_harmonics()
            
            self.vector = vector
        else:
            print('ERROR: incompatible basis and vector input. Check that both have same length.')
        
    def redefine_vector(self,vector):

        '''
        Update vector definition

        *args*:

            - **vector**: numpy array of complex float, same length as self.vector

        ***
        '''

        try:
            self.vector[:] = vector
        except ValueError:
            print('Error: Input vector is not of the same shape as original selection. Please check input vector.')
            
        
    
    def find_harmonics(self):
        '''
        Create a pointer array of basis indices and the associated spherical harmonics, as well as
        aa more convenient vector form of the projections themselves, as lists of complex float
        
        *return*:

            - **all_lm**: list of int, l,m pairs of all spherical harmonics relevant to calculation
            
            - **lm_pointers**: list of int, pointer indices relating each basis orbital projection to the 
            lm pairs in *all_lm*
            
            - **projectors**: list of arrays of complex float, providing the complex projection of basis
            onto the related spherical harmonics
        
        ***
        '''
        all_lm = []
        lm_pointers = []
        projectors = []
        for o in self.basis:
            proj_pointers = np.zeros(len(o.proj))
            proj_vals = np.zeros(len(o.proj),dtype=complex)
            for oi in range(len(o.proj)):
                proj_vals[oi] = o.proj[oi][0]+1.0j*o.proj[oi][1]
                lm = np.array([o.proj[oi][2],o.proj[oi][3]]).astype(int)
                try:
                    d_lm = np.linalg.norm(np.array([lm_ii - lm for lm_ii in all_lm]),axis=1)
                    if d_lm.min()==0:
                        index = np.where(d_lm==0)[0][0]
                        proj_pointers[oi]=index
                    else:
                        all_lm.append(lm)
                        proj_pointers[oi] = len(all_lm)-1
                except ValueError:
                    all_lm.append(lm)
                    proj_pointers[0] = 0
            lm_pointers.append(list(proj_pointers.astype(int)))
            projectors.append(proj_vals)
        return all_lm,lm_pointers,projectors
        
            
    def find_centres(self):
        '''
        Create a Pointer array of basis indices and the centres of these basis orbitals.
        
        *return*:

            - **all_centres**: list of numpy array of length 3, indicating unique positions in the basis set
            
            - **centre_pointers**: list of int, indicating the indices of position array, associated with the
            location of the related orbital in real space.
        
        '''
        all_centres = []
        centre_pointers = []
        for o in self.basis:
            centre = o.pos
            try:
                d_centres = np.linalg.norm(np.array([centre-ac for ac in all_centres]),axis=1)
                if d_centres.min()==0.0:
                    index = np.where(d_centres==0)[0][0]
                    centre_pointers.append(index)
                else:
                    all_centres.append(centre)
                    centre_pointers.append(len(all_centres)-1)
            except ValueError:
                all_centres.append(centre)
                centre_pointers.append(0)
            
        return all_centres,centre_pointers
    
    
    def calc_Ylm(self,th,ph):

        '''
        Calculate all spherical harmonics needed for present calculation
        
        *return*:

            - numpy array of complex float, of shape (len(self.harmonics),len(th))
        
        ***
        '''
        return np.array([Ylm.Y(int(lm[0]),int(lm[1]),th,ph) for lm in self.harmonics])
        
        
                
            
    def triangulate_wavefunction(self,n,plotting=True,ax=None):

        '''
        Plot the wavefunction stored in the class attributes as self.vector as a projection
        over the basis of spherical harmonics. The radial wavefunctions are not explicitly included,
        in the event of multiple basis atom sites, the length scale is set by the mean interatomic 
        distance. The wavefunction phase is encoded in the colourscale of the mesh plot. The user
        sets the smoothness of the orbital projection by the integer argument *n*
        
        *args*:

            - **n**: int, number of angles in the mesh: Theta from 0 to pi is divided 2n times, and
            Phi from 0 to 2pi is divided 4n times
            
        *kwargs*:

            - **plotting**: boolean, turn on/off to display plot
            
            - **ax**: matplotlib Axes, for plotting on existing plot
            
            
        *return*:

            - **vertices**: numpy array of float, shape (len(centres), len(th)*len(ph), 3) locations of vertices
            
            - **triangulations**: numpy array of int, indicating the vertices connecting each surface patch
            
            - **colours**: numpy array of float, of shape (len(centres),len(triangles)) encoding the orbital phase for each surface patch of the plotting
        
            - **ax**: matplotlib Axes, for further modifications
        
        ***
        '''
        th,ph = make_angle_mesh(n)
        all_Ylm = self.calc_Ylm(th,ph)
        
        if len(self.centres)>1:
            ad = 0.5*np.mean(np.array([np.linalg.norm(self.centres[i]-self.centres[j]) for i in range(len(self.centres)) for j in range(i,len(self.centres))]))
        else:
            ad = 4.0
            
        ncentres = len(self.centres)
        vertices = np.zeros((ncentres,len(th),3))
        radii = np.zeros((ncentres,len(th)),dtype=complex)
        triangulations = mtri.Triangulation(th,ph)
        colours = []
        
        for bi in range(len(self.basis)):

            radii[self.centre_pointer[bi],:] += np.sum(np.array([all_Ylm[self.harmonic_pointer[bi][j]]*self.vector[bi]*self.projections[bi][j] for j in range(len(self.harmonic_pointer[bi]))]),axis=0)
        
        rescale = ad/np.mean(abs(radii)**2)
        for ni in range(ncentres):
            vertices[ni,:,:]+=rescale*np.array([abs(radii[ni])**2*np.cos(ph)*np.sin(th),abs(radii[ni])**2*np.sin(th)*np.sin(ph),abs(radii[ni])**2*np.cos(th)]).T
            colours.append(col_phase(radii[ni,triangulations.triangles][:,1]))
            vertices[ni,:]+=self.centres[ni]
            
        colours = np.array(colours)
        if plotting:
            
            _,ax = self.plot_wavefunction(vertices,triangulations,colours,plot_ax=ax)
            
        return vertices,triangulations,colours,ax
    
    
    def plot_wavefunction(self,vertices,triangulations,colours,plot_ax = None,cbar_ax= None):
        '''
        Plotting function, for visualizing orbitals.
        
        *args*:

            - **vertices**: numpy array of float, shape (len(centres), len(th)*len(ph), 3) locations of vertices
            
            - **triangulations**: numpy array of int, indicating the vertices connecting each surface patch
            
            - **colours**: numpy array of float, of shape (len(centres),len(triangles)) encoding the orbital phase for each surface patch of the plotting
        
            - **plot_ax**: matplotlib Axes, for plotting on existing axes
            
            - **cbar_ax**: matplotlib Axes, for use in drawing colourbar
            
        *return*:

            - **plots**: list of plotted surfaces
            
            - **plot_ax**: matplotlib Axes, for further modifications
            
        ***
        '''
        ncentres = len(self.centres)
        plots = []
        if plot_ax is None:
            fig = plt.figure()
            plot_ax = fig.add_subplot(111,projection='3d')

        for ni in range(ncentres):
            plots.append(plot_ax.plot_trisurf(vertices[ni,:,0],vertices[ni,:,1],vertices[ni,:,2],triangles=triangulations.triangles,cmap=cm.hsv,antialiased=True,edgecolors='w',linewidth=0.2))
            plots[-1].set_array(colours[ni])
            plots[-1].set_clim(-np.pi,np.pi)
            
        plot_ax.set_xlabel('X')
        plot_ax.set_ylabel('Y')
        plot_ax.set_zlabel('Z')
        plt.colorbar(plots[-1],ax=plot_ax,cax=cbar_ax)
        
        return plots,plot_ax
            

def make_angle_mesh(n):
    '''
    Quick utility function for generating an angular mesh over spherical surface
    
    *args*:

        - **n**: int, number of divisions of the angular space
        
    *return*:

        - **th**: numpy array of 2n float from 0 to pi
        
        - **ph**: numpy array of 4n float from 0 to 2pi
    
    ***
    '''
    th = np.linspace(0,np.pi,2*n)
    ph = np.linspace(0,2*np.pi,4*n)
    th,ph = np.meshgrid(th,ph)
    th,ph = th.flatten(),ph.flatten()
    return th,ph


def col_phase(vals):

    '''
    Define the phase of a complex number
    
    *args*:

        - **vals**: complex float, or numpy array of complex float
        
    *return*:

        - float, or numpy array of float of same shape as vals, from -pi to pi
    
    ***
    '''
    x,y=np.real(vals),np.imag(vals)
    return np.arctan2(y,x)


def rephase_wavefunctions(vecs,index=-1):

    '''
    The wavefunction at different k-points can choose an arbitrary phase, as can 
    a subspace of degenerate eigenstates. As such, it is often advisable to choose
    a global phase definition when comparing several different vectors. The user here
    passes a set of vectors, and they are rephased. The user has the option of specifying
    which basis index they would like to set the phasing. It is essential however that the
    projection onto at least one basis element is non-zero over the entire set  of vectors 
    for this rephasing to work.
    
    *args*:

        - **vecs**: numpy array of complex float, ordered as rows:vector index, columns: basis index
        
    *kwargs*:

        - **index**: int, optional choice of basis phase selection
        
    *return*:

        - **rephase**: numpy array of complex float of same shape as *vecs*
    
    ***
    '''
    
    rephase = np.copy(vecs)
    if index>-1:
        #check that user has selected a viable phase choice
        if abs(vecs[:,index]).min()<1e-10:
            print('Warning, the chosen basis index is invalid. Please make another selection.\n')
            print('Finite projection onto the basis element of choice must be finite. If you are\n')
            print('unsure, the computer can attempt to make a viable selection in the absence of\n')
            print('an indicated basis index.')
            return rephase
    
    else:
        min_projs = np.array([abs(vecs[:,i]).min() for i in range(np.shape(vecs)[0])])
        index = np.where(min_projs>0)[0][0]
    phase_factors = np.conj(vecs[:,index])/abs(vecs[:,index])
    rephase = np.einsum('ij,i->ij',rephase,phase_factors)
    
    return rephase
    
            
        
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#Created on Sat Nov 18 11:15:35 2017

#@author: ryanday

#MIT License

#Copyright (c) 2018 Ryan Patrick Day

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.


import numpy as np

import sys
if sys.version_info<(3,0):
    raise ('This software requires Python 3.0 or higher. Please update your Python installation before proceeding')
else:
    import chinook.orbital as olib
    import chinook.TB_lib as TBlib
    import chinook.slab as slib
    import chinook.klib as klib




###Build Basis
def gen_basis(basis):
    '''
    Generate a list of orbital objects as the input basis for a tight-binding model.
    User passes a basis dictionary, function returns a modified version of this
    same dictionary, with the list of orbitals now appended as the *'bulk'* entry
    
    *args*:

        - **basis**--dictionary with keys:
            
            - *'atoms'*: list of integer, indices for distinct atoms, 
            
            - *'Z'*: dictionary of integer: *'atom'*:element (integer) pairs
            
            - *'orbs'*: list of lists of string, for each atom containing the
            orbital labels (usually in conventional nlxx format)), 
            
            - *'pos'*: list of numpy arrays of length 3 float indicating
            positions of the atoms in direct Angstrom units, 
            
            - optional keys: 
    
                - *'orient'*: list, one entry for each atom, indicating a
                local rotation of the indicated atom, various formats accepted; 
                for more details, c.f. **chinook.orbital.py**
                
                - *'spin'*: dictionary of spin information:
                
                    - *'bool'*: boolean, double basis into spinor basis, 
                    
                    - *'soc'*: boolean, include spin-orbit coupling
                    
                    - *'lam'*: dictionary of SOC constants, integer:float
                    pairs for atoms in *'atoms'* list, and lambda_SOC in eV

                    
    *return*:

        - **basis** dictionary, modified to include the **bulk** list of orbital
        objects

        
    ***
    '''
	
    bulk_basis = []
    
    required = ['atoms','orbs','pos','Z']
    all_present = recur_product([ri in basis.keys() for ri in required])
    do_orient = 'orient' in basis.keys()
    if do_orient:
        for a in range(len(basis['atoms'])):
            if len(basis['orient'][a])==1:
                basis['orient'][a] = [basis['orient'][a] for i in range(len(basis['orbs'][a]))]
            elif len(basis['orient'][a])<len(basis['orbs'][a]):
                raise ValueError ('ORIENT ERROR: pass either 1 orientation per orbital for a given atom, or a single orientation for all orbitals on atom')
                return None
    if not all_present:
        raise ValueError ('BASIS GENERATION ERROR!!!! Ensure atoms, atomic numbers, orbitals, positions are all passed to gen_basis in the basis dictionary. See gen_basis.__doc__ for details.')
        return None
    else:
        for a in list(enumerate(basis['atoms'])):
            for o in list(enumerate(basis['orbs'][a[0]])):
                if do_orient:
            
                    bulk_basis.append(olib.orbital(a[1],len(bulk_basis),o[1],basis['pos'][a[0]],basis['Z'][a[1]],orient=basis['orient'][a[0]][o[0]]))
                else:
                    bulk_basis.append(olib.orbital(a[1],len(bulk_basis),o[1],basis['pos'][a[0]],basis['Z'][a[1]]))
    if 'spin' in basis.keys():
        if basis['spin']['bool']:
            bulk_basis = olib.spin_double(bulk_basis,basis['spin']['lam'])  
    basis['bulk'] = bulk_basis
    
    return basis

def gen_K(Kdic):    
    '''
    Generate k-path for TB model to be diagonalized along.
    
    *args*: 

        - **Kdic**: dictionary for generation of kpath with:
        
            - *'type'*: string 'A' (absolute) or 'F' (fractional) units 
        
            - *'avec'*: numpy array of 3x3 float lattice vectors
        
            - *'pts'*: list of len3 array indicating the high-symmetry points 
            along the path of interest
        
            - *'grain'*: int, number of points between *each* element of *'pts'*
        
            optional: 
        
                - *'labels'*:list of strings with same length as *'pts'*, giving
                plotting labels for the kpath
            
    *return*:

        **Kobj**: K-object including necessary attributes to be read by the **TB_model**
    
    
    '''
    
    if 'labels' not in Kdic.keys():
        Kdic['labels'] = ['K{:d}'.format(i) for i in range(len(Kdic['pts']))]
    required = ['type','pts','grain']
    if not recur_product([ri in Kdic.keys() for ri in required]):
        raise KeyError('Invalid K-dictionary format. See documentation for gen_K to ensure all required arguments are passed in k-dictionary')
        return None
    if Kdic['type'] == 'F' and 'avec' not in Kdic.keys():
        raise KeyError('Invalid K-dictionary format. Must pass lattice vectors for fractional coordinates')
        return None
    else:
        if Kdic['type']=='F':
            B = klib.bvectors(Kdic['avec'])
            klist = [np.dot(k,B) for k in Kdic['pts']]
        elif Kdic['type']=='A':
            klist = [k for k in Kdic['pts']]
        else:
            klist = []
            print('You have not entered a valid K path. Proceed with caution.')
        Kobj = klib.kpath(klist,Kdic['grain'],Kdic['labels'])
    
        return Kobj


###Built Tight Binding Model
def gen_TB(basis_dict,hamiltonian_dict,Kobj=None,slab_dict=None):
    '''
    Build a Tight-Binding Model using the user-input dictionaries
    
    *args*:

        - **basis_dict**: dictionary, including the *'bulk'* key value pair
        generated by **gen_basis**
        
        - **hamiltonian_dict**: dictionary,
        
            - *'spin'*: same dictionary as passed to **gen_basis** 
            
            - *'type'*: string, Hamiltonian type--'list' (list of matrix elements),
            'SK' (Slater-Koster dictionaries, requires also a 'V' and 'avec' entry),
            'txt' (textfile, requires a 'filename' key as well)
            
            - *'cutoff'*: float, cutoff hopping distance
            
            - *'renorm'*: float, renormalization factor
            
            - *'offset'*: float, offset of chemical potential
            
            - *'tol'*: float, minimum matrix element tolerance
                
        - **Kobj**: optional, standard K-object, as generated by **gen_K**
        
        - **slab_dict**: dictionary for slab generation
        
            - *'avec'*: numpy array of 3x3 float, lattice vectors
            
            - *'miller'*: numpy array of 3 integers, indicating the Miller 
            index of the surface normal in units of lattice vectors
            
            - *'fine'*: fine adjustment of the slab thickness, tuple of two 
            numeric to get desired termination correct (for e.g. inversion symmetry)
            
            - *'thick'*: integer approximate number of unit cells in the
            slab (will not be exact, depending on the fine, and termination
            
            - *'vac'*: int size of the vacuum buffer -- must be larger than
            the largest hopping length to ensure no coupling of slabs
            
            - *'termination'*: tuple of 2 integers: atom indices which 
            terminate the top and bottom of the slab
        

    *return*:

        **TB_model**: tight-binding object, as defined in **chinook.TB_lib.py**
    
    '''
    if 'spin' not in hamiltonian_dict.keys():
        # if omitted, assume no spin-degree of freedom desired
        print('No spin-information entered, assuming no spin-degree of freedom in the following. See build_lib.py for details if spin is desired.')
        hamiltonian_dict['spin']={'bool':False}
    required = ['type','cutoff','renorm','offset','tol']
    
    if not recur_product([ri in hamiltonian_dict.keys() for ri in required]):
        raise ValueError ('Ensure all requisite arguments passed in the Hamiltonian dictionary. see gen_TB documentation for details.')
        return None
    else:
        if hamiltonian_dict['type']=='SK' and ('V' not in hamiltonian_dict.keys() or 'avec' not in hamiltonian_dict.keys()):
            raise ValueError ('PLEASE INCLUDE THE DICTIONARY OF Slater-Koster elements as "V" in the Hamiltonian dictionary, and lattice vectors "avec" as numpy array of 3x3 float.')
            return None
        elif hamiltonian_dict['type']=='txt' and 'filename' not in hamiltonian_dict.keys():
            raise ValueError ('No "filename" included in Hamiltonian dictionary keys for text-based Hamiltonian entry.')
            return None
        elif hamiltonian_dict['type']=='list' and 'list' not in hamiltonian_dict.keys():
            raise KeyError ('No "list" included in Hamiltonian dictionary keys for list-based Hamiltonian entry.')
            return None
        else:
            if type(slab_dict)==dict:
                if hamiltonian_dict['spin']['bool']:
                    basis_dict['bulk'] = basis_dict['bulk'][:int(len(basis_dict['bulk'])/2)]
                    Hspin = True
                    hamiltonian_dict['spin']['bool'] = False #temporarily forestall incorporation of spin
                else:
                    Hspin=False
            TB = TBlib.TB_model(basis_dict['bulk'],hamiltonian_dict,Kobj)
            if type(slab_dict)==dict:
                slab_dict['TB'] = TB
                print('running bulk_to_slab now')
                TB,slab_H,Rmat = slib.bulk_to_slab(slab_dict) 
                if Hspin:
                    TB.basis = olib.spin_double(list(TB.basis),basis_dict['spin']['lam']) 
        
                hamiltonian_dict['type']='list'
                hamiltonian_dict['list'] = slab_H
                
                hamiltonian_dict['avec'] = TB.avec
                hamiltonian_dict['spin']['bool']=Hspin
        
                TB.mat_els = TB.build_ham(hamiltonian_dict)
        return TB
    

def recur_product(elements):
    '''
    Utility function: Recursive evaluation of the product of all elements in a list
    
    *args*:

        - **elements**: list of numeric type
    
    *return*:
    
        - product of all elements of **elements**
    
    ***
    '''
    if len(elements)==1:
        return elements[0]
    else:
        return elements[0]*recur_product(elements[1:])






#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#Created on Mon Nov 13 18:28:22 2017
#@author: ryanday
#MIT License

#Copyright (c) 2018 Ryan Patrick Day

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.


#For now, given a set of pts--numpy array 3xN and a grain --# points between each of the N pts,
#this will generate a Kobject, with labels, grain, and other attributes useful for plotting
#In the future it would be nice to generate a klist based on an xcrysden calculation for example.



import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D


me= 9.11e-31
hb = 6.626e-34/(2*np.pi)
q = 1.602e-19
A=1.0e-10

class kpath:
    
    '''
    Momentum object, defining a path in reciprocal space, for use in defining the 
    Hamiltonian at different points in the Brillouin zone.
    ***
    '''
    
    def __init__(self,pts,grain=None,labels=None):
        
        '''
        Initialize the momentum object
        
        *args*:
            - **pts**: list of len(3) numpy array of float: the endpoints of the 
            path in k-space
            
        *kwargs*:
            - **grain**: int, optional, indicating number of points between each point 
            in **pts**
            
            - **labels**: list of strings, optional, labels for plotting bandstructure
            along the kpath
        ***
        '''
        
        if type(pts)==list:
            self.pts = pts
            self.grain = grain
            self.kpts = [] #will be the actual vector k points for calculation
            self.kcut = [] #will be a 1D array for plotting purposes
            self.kcut_brk = []
            self.ind_brk = [0]
            self.labels=labels
            self.points()
        elif type(pts)==np.ndarray:
            self.kpts = pts #if we are just feeding in an array of kpoints
            
    def points(self):
        
        '''
        Use the endpoints of kpath defined in **kpath.pts** to create numpy array
        of len(3) float which cover the entire path, based on method by I.S. Elfimov.
        
        *return*:
            - **kpath.kpts**: numpy array of float, len(**kpath.pts**)(1+**kpath.grain**) by 3
        
        ***
        '''
        
        cutst = 0.0
        self.kcut_brk.append(cutst)
        self.kpts = [(self.pts[0])]
        self.kcut.append(cutst)
        for p in range(len(self.pts)-1):
            tempvec = np.array(self.pts[p+1])-np.array(self.pts[p])
            vecstep = tempvec/self.grain

            for i in range(1,self.grain+1):
                self.kpts.append(list(self.pts[p]+vecstep*float(i)))
                self.kcut.append(cutst+np.linalg.norm(vecstep)*float(i))
            cutst = self.kcut[-1]
            self.kcut_brk.append(cutst)
            self.ind_brk.append(len(self.kcut))
        self.kpts = np.array(self.kpts)
        return self.kpts
    
    
def bvectors(a_vec):
    '''
    Define the reciprocal lattice vectors corresponding to the direct lattice 
    in real space
    
    *args*:
        - **a_vec**: numpy array of 3x3 float, lattice vectors
        
    *return*:
        - **b_vec**: numpy array of 3x3 float, reciprocal lattice vectors
       
    ***
    '''
    b_vec = 2*np.pi*np.array([(np.cross(a_vec[1],a_vec[2])/np.dot(a_vec[0],np.cross(a_vec[1],a_vec[2]))),(np.cross(a_vec[2],a_vec[0])/np.dot(a_vec[1],np.cross(a_vec[2],a_vec[0]))),(np.cross(a_vec[0],a_vec[1])/np.dot(a_vec[2],np.cross(a_vec[0],a_vec[1])))])
    return b_vec


def b_zone(a_vec,N,show=False):
    '''
    Generate a mesh of points over the Brillouin zone. 
    Each of the cardinal axes are divided by the same number of points 
    (so points are not necessarily evenly spaced along each axis).
    
    *args*:
        - **a_vec**: numpy array of size 3x3 float
        
        - **N**: int mesh density
   
    *kwargs*:     
        - **show**: boolean for optional plotting of the mesh points
        
    *return*:
        - **m_pts**: numpy array of mesh points (float), shape (len(m_pts),3)
    '''
    
    b_vec = bvectors(a_vec)
    rlpts = np.dot(region(1),b_vec)    
    bz_mesh_raw = raw_mesh(rlpts,N)
    bz_mesh = mesh_reduce(rlpts,bz_mesh_raw)
    
    if show:
        plt_pts(bz_mesh)

    return bz_mesh



def region(num):
    '''
    Generate a symmetric grid of points in number of lattice vectors.
    
    *args*:
        - **num**: int, grid will have size 2 num+1 in each direction
        
    *return*:
        - numpy array of size ((2 num+1)**3,3) with centre value of 
        first entry of (-num,-num,-num),...,(0,0,0),...,(num,num,num)
        
    ***
    '''
    num_symm = 2*num+1
    return np.array([[int(i/num_symm**2)-num,int(i/num_symm)%num_symm-num,i%num_symm-num] for i in range((num_symm)**3)])



def raw_mesh(blatt,N):
    
    '''
    Define a mesh of points filling the region of k-space bounded by the set
    of reciprocal lattice points generated by *bvectors*.
    These will be further reduced by *mesh_reduce* to find points which
    are within the first-Brillouin zone
    
    *args*:
        - **blatt**: numpy array of 27x3 float
        
        - **N**: int, or iterable of len 3, defines a coarse estimation
        of number of k-points
        
    *return*:
        - **mesh**: numpy array of mesh points, size set roughly by N
        
    ***
    '''
    limits = np.array([[blatt[:,ii].min(),blatt[:,ii].max()] for ii in range(3)])
    if type(N)==int:
        L_max = (limits[:,1]-limits[:,0]).min()
        dk = [(L_max/N)]*3
    elif type(N)==tuple or type(N)==list or type(N)==np.ndarray:
        if len(N)==3:
            dk = [(limits[i,1]-limits[i,0])/N[i] for i in range(3)]
        else:
            print('ERROR: Must pass 3 numbers if passing a list/tuple/array')
            return []
        
    else:
        print('Invalid K division type. Pass only integer or list/tuple/array of 3 integer values')
        return []
    x,y,z = np.arange(limits[0,0],limits[0,1]+dk[0],dk[0]),np.arange(limits[1,0],limits[1,1]+dk[1],dk[1]),np.arange(limits[2,0],limits[2,1]+dk[2],dk[2])
    Y,Z,X = np.meshgrid(y,z,x) #This ordering for the meshgrid allows us to define 1D array with register looping over x, then y, then z. Not sure why this order gives that output...
    X,Y,Z = X.flatten(),Y.flatten(),Z.flatten()
    Kpts = np.array([[X[i],Y[i],Z[i]] for i in range(len(X))])
    return Kpts

def mesh_reduce(blatt,mesh,inds=False):
    '''
    Determine and select only k-points corresponding to the first Brillouin
    zone, by simply classifying points on the basis
    of whether or not the closest lattice point is the origin. 
    By construction, the origin is index 13 of the blatt. 
    If it is not, return error. Option to take only the indices of
    the mesh which we want, rather than the actual array points
    --this is relevant for tetrahedral interpolation methods
    
    *args*:
        - **blatt**: numpy array of len(27,3), nearest reciprocal lattice vector points
        
        - **mesh**: numpy array of (N,3) float, defining a mesh of k points, before
        being reduced to contain only the Brillouin zone.
        
    *kwargs*: 
        - **inds**: option to pass a list of bool, indicating the 
        indices one wants to keep, instead of autogenerating the mesh
    
    *return*:
        - **bz_pts**: numpy array of (M,3) float, Brillouin zone points
    
    '''
    bz_pts = []
    if np.linalg.norm(blatt[13])>0:       
        print('FORMAT ERROR: invalid Reciprocal Lattice Point array passed. Please use np.dot(region(1),b_vec) to generate these points')
        return []
    else:
        for m in list(enumerate(mesh)):
            dv = np.linalg.norm(np.around(m[1]-blatt,4),axis=1)
            if (13 in np.where(dv==dv.min())[0]):
                bz_pts.append(m[not inds])


        return np.array(bz_pts)
    

def plt_pts(pts):
    '''
    Plot an array of points iin 3D
    
    *args*:
        - **pts**: numpy array shape N x 3
    
    '''
    fig = plt.figure()
    ax = fig.add_subplot(111,projection='3d')
    ax.scatter(pts[:,0],pts[:,1],pts[:,2])
 
    
    
def kmesh(ang,X,Y,kz,Vo=None,hv=None,W=None):
    '''
    Take a mesh of kx and ky with fixed kz and generate a Nx3 array of points
    which rotates the mesh about the z axis by **ang**. N is the flattened shape
    of **X** and **Y**.
    
    *args*:
        - **ang**: float, angle of rotation
        
        - **X**: numpy array of float, one coordinate of meshgrid
        
        - **Y**: numpy array of float, second coordinate of meshgrid
        
        - **kz**: float, third dimension of momentum path, fixed
        
        
    *kwargs*:
        - **Vo**: float, parameter necessary for inclusion of inner potential
        
        - **hv**: float, photon energy, to be used if **Vo** also included, for evaluating kz
        
        - **W**: float, work function
        
    *return*:
        - **k_arr**: numpy array of shape Nx3 float, rotated  kpoint array.
        
        - **ph**: numpy array of N float, angles of the in-plane momentum
        points, before rotation.
    
    ***    
    '''
            
    kp = np.sqrt(X**2+Y**2)
    ph = np.arctan2(Y,X)
    if abs(ang)>0.0:
        X = kp*np.cos(ph-ang)
        Y = kp*np.sin(ph-ang)
    try:
        Zeff = kz_kpt(hv,kp,W,Vo)
        print('Inner potential detected, overriding kz input.')
    except TypeError:    
        Zeff = kz*np.ones(np.shape(X))
    
    ph = np.reshape(ph,np.shape(X)[0]*np.shape(X)[1])
    k_arr = np.reshape(np.array([X,Y,Zeff]),(3,np.shape(X)[0]*np.shape(X)[1])).T
    return k_arr,ph


def kmesh_hv(ang,x,y,hv,Vo=None):
    '''
    Take a mesh of kx and ky with fixed kz and generate a Nx3 array of points
    which rotates the mesh about the z axis by **ang**. N is the flattened shape
    of **X** and **Y**.
    
    *args*:
        - **ang**: float, angle of rotation
        
        - **X**: numpy array of float, one coordinate of meshgrid
        
        - **Y**: numpy array of float, second coordinate of meshgrid
        
        - **kz**: float, third dimension of momentum path, fixed
        
        
    *kwargs*:
        - **Vo**: float, parameter necessary for inclusion of inner potential
        
        - **hv**: float, photon energy, to be used if **Vo** also included, for evaluating kz
        
        - **W**: float, work function
        
    *return*:
        - **k_arr**: numpy array of shape Nx3 float, rotated  kpoint array.
        
        - **ph**: numpy array of N float, angles of the in-plane momentum
        points, before rotation.
    
    ***    
    '''
            

    
    if len(x)==1:

        X,HV = np.meshgrid(y,hv)
        Y = x*np.ones(np.shape(X))
    elif len(y)==1:
        X,HV = np.meshgrid(x,hv)
        Y = y*np.ones(np.shape(X))
        
        
    ph = np.arctan2(Y,X)
    KP = np.sqrt(X**2+Y**2)    

    if abs(ang)>0.0:
        x = KP*np.cos(ph-ang)
        y = KP*np.sin(ph-ang)
    Z = kz_kpt(HV,KP,V=Vo)
    
    ph = np.reshape(ph,np.shape(X)[0]*np.shape(X)[1])
    k_arr = np.reshape(np.array([X,Y,Z]),(3,np.shape(X)[0]*np.shape(X)[1])).T
    return k_arr,ph


def kz_kpt(hv,kpt,W=0,V=0):
    
    '''
    Extract the kz associated with a given in-plane momentum, photon energy, 
    work function and inner potential
    
    *args*:
        - **hv**: float, photon energ in eV
        
        - **kpt**: float, in plane momentum, inverse Angstrom
        
        - **W**: float, work function in eV
        
        - **V**: float, inner potential in eV
        
    *return*:
        - **kz**: float, out of plane momentum, inverse Angstrom
    ***
    '''
    kn = np.sqrt(2*me/hb**2*(hv-W)*q)*A
    
    kz= np.sqrt(2*me/hb**2*((hv-W)*(1-(kpt/kn)**2)+V)*q)*A
    
    return kz

if __name__=="__main__":
    a = 1.78
    av = np.array([[a,0,a],[0,a,a],[a,a,0]])
#    av = np.array([[np.sqrt(3)*a/2,a/2,0],[np.sqrt(3)*a/2,-a/2,0],[0,0,2*c]])
    N = 20
    bz = b_zone(av,N,True)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#Created on Sun Jun 18 15:46:21 2017

#@author: ryanday
#MIT License

#Copyright (c) 2018 Ryan Patrick Day

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

import numpy as np
from math import factorial



projdict={"0":np.array([[1.0,0.0,0.0,0.0]]),
               "1x":np.array([[-np.sqrt(0.5),0.0,1,1],[np.sqrt(0.5),0.0,1,-1]]),"1y":np.array([[0.0,np.sqrt(0.5),1,1],[0,np.sqrt(0.5),1,-1]]),"1z":np.array([[1,0,1,0]]),
                "2xy":np.array([[0.0,-np.sqrt(0.5),2,2],[0.0,np.sqrt(0.5),2,-2]]),"2yz":np.array([[0.0,np.sqrt(0.5),2,1],[0.0,np.sqrt(0.5),2,-1]]),
                "2xz":np.array([[-np.sqrt(0.5),0,2,1],[np.sqrt(0.5),0,2,-1]]),"2ZR":np.array([[1.0,0.0,2,0]]),"2XY":np.array([[np.sqrt(0.5),0,2,2],[np.sqrt(0.5),0,2,-2]]),
                "3z3":np.array([[1.0,0.0,3,0]]),"3xz2":np.array([[np.sqrt(0.5),0,3,-1],[-np.sqrt(0.5),0,3,1]]),
                "3yz2":np.array([[0,np.sqrt(0.5),3,-1],[0,np.sqrt(0.5),3,1]]),"3xzy":np.array([[0,-np.sqrt(0.5),3,2],[0,np.sqrt(0.5),3,-2]]),
                "3zXY":np.array([[np.sqrt(0.5),0,3,2],[np.sqrt(0.5),0,3,-2]]),"3xXY":np.array([[-np.sqrt(0.5),0,3,3],[np.sqrt(0.5),0,3,-3]]),
                "3yXY":np.array([[0,np.sqrt(0.5),3,3],[0,np.sqrt(0.5),3,-3]])}
          

def Y(l,m,theta,phi):
    '''
    Spherical harmonics, defined here up to l = 4. This allows for photoemission from
    initial states up to and including f-electrons (final states can be d- or g- like).
    Can be vectorized with numpy.vectorize() to allow array-like input
    
    *args*:

        - **l**: int orbital angular momentum, up to l=4 supported
        
        - **m**: int, azimuthal angular momentum |m|<=l
        
        - **theta**: float, angle in spherical coordinates, radian measured from the z-axis [0,pi]
        
        - **phi**: float, angle in spherical coordinates, radian measured from the x-axis [0,2pi]

    *return*:

        - complex float, value of spherical harmonic evaluated at theta,phi
    
    '''
    
    
    if l == 0:
        if m==0:
            return 0.5*np.sqrt(1.0/np.pi)*value_one(theta,phi)
        else:
            return 0.0 
               
    elif l == 1:
        if abs(m) == 1:
            return -np.sign(m)*0.5*np.sqrt(3.0/(2*np.pi))*np.exp(m*1.0j*phi)*np.sin(theta)
        elif m == 0:
            return 0.5*np.sqrt(3.0/np.pi)*np.cos(theta) 
        else:
            return 0.0

    elif l == 2:
        if abs(m) == 2:
            return 0.25*np.sqrt(15.0/(2*np.pi))*np.exp(m*1.0j*phi)*np.sin(theta)**2
        elif abs(m) == 1:
            return -np.sign(m)*0.5*np.sqrt(15.0/(2*np.pi))*np.exp(m*1.0j*phi)*np.sin(theta)*np.cos(theta)
        elif m==0:
            return 0.25*np.sqrt(5/np.pi)*(3*np.cos(theta)**2-1)
        else:
            return 0.0

    elif l == 3:
        if abs(m) == 3:
            return -np.sign(m)*1.0/8*np.sqrt(35/np.pi)*np.exp(m*1.0j*phi)*np.sin(theta)**3
        elif abs(m) == 2:
            return 1.0/4*np.sqrt(105/(2*np.pi))*np.exp(m*1.0j*phi)*np.sin(theta)**2*np.cos(theta)
        elif abs(m) == 1:
            return -np.sign(m)*1.0/8*np.sqrt(21/np.pi)*np.exp(m*1.0j*phi)*np.sin(theta)*(5*np.cos(theta)**2-1)
        elif m == 0:
            return 1.0/4*np.sqrt(7/np.pi)*(5*np.cos(theta)**3-3*np.cos(theta))
        else:
            return 0.0

    elif l == 4:
        if abs(m) == 4:
            return 3/16.*np.sqrt(35./2/np.pi)*np.sin(theta)**4*np.exp(m*1.0j*phi)
        elif abs(m) == 3:
            return -np.sign(m)*3/8.*np.sqrt(35/np.pi)*np.sin(theta)**3*np.cos(theta)*np.exp(m*1.0j*phi)
        elif abs(m) == 2:
            return 3./8.*np.sqrt(5/2/np.pi)*np.sin(theta)**2*(7*np.cos(theta)**2-1.0)*np.exp(m*1.0j*phi) 
        elif abs(m) == 1:
            return -np.sign(m)*3./8*np.sqrt(5/np.pi)*np.sin(theta)*(7*np.cos(theta)**3-3*np.cos(theta))*np.exp(m*1.0j*phi)
        elif m == 0:
            return 3./16.*np.sqrt(1./np.pi)*(35.*np.cos(theta)**4 - 30.*np.cos(theta)**2 + 3.0)
        else:
            return 0.0

    else:
        return 0.0
    
    
def value_one(theta,phi):
    '''
    Flexible generation of the number 1.0, in either float or array format

    *args*:

        - **theta**: float or numpy array of float

        - **phi**: float or numpy array of float

    *return*:

        **out**: float or numpy array of float, evaluated to 1.0, of same shape and type
        as **theta**, **phi**

    ***
    '''
    if type(theta)==np.ndarray:
        out =np.ones(np.shape(theta))
    if type(phi)==np.ndarray:
        out = np.ones(np.shape(phi))
    elif type(theta)!=np.ndarray and type(phi)!=np.ndarray:
        out = 1.0
    return out
        

def binom(a,b):
    '''
    Binomial coefficient for 'a choose b'

    *args*:

        - **a**: int, positive

        - **b**: int, positive

    *return*:

        - float, binomial coefficient

    ***
    '''
    return factorial(a+b)/float(factorial(a-b)*factorial(b))

def laguerre(x,l,j):
    '''
    Laguerre polynomial of order l, degree j, evaluated over x

    *args*:

        - **x**: float or numpy array of float, input

        - **l**: int, order of polynomial

        - **j**: int, degree of polynomial


    *return*:

        - **laguerre_output**: float or numpy array of float, shape as input **x**

    ***
    '''
    laguerre_output = sum([((-1)**i)*(binom(l+j,j-i)*x**i/float(factorial(i))) for i in range(j+1)])
    return laguerre_output


def gaunt(l,m,dl,dm):
    '''
    I prefer to avoid using the sympy library where possible, for speed reasons. These are the explicitly defined
    Gaunt coefficients required for dipole-allowed transitions (dl = +/-1) for arbitrary m,l and dm
    These have been tested against the sympy package to confirm numerical accuracy for all l,m possible
    up to l=5. This function is equivalent, for the subset of dm, dl allowed to
    sympy.physics.wigner.gaunt(l,1,l+dl,m,dm,-(m+dm))
    
    *args*:
        
        - **l**: int orbital angular momentum quantum number
        
        - **m**: int azimuthal angular momentum quantum number
        
        - **dl**: int change in l (+/-1)
        
        - **dm**: int change in azimuthal angular momentum (-1,0,1)
    
    *return*:

        - float Gaunt coefficient

    ***
    '''
    try:
        if abs(m + dm)<=(l+dl):
            if dl==1:
                if dm == 1:
                    return (-1.)**(m+1)*np.sqrt(3*(l+m+2)*(l+m+1)/(8*np.pi*(2*l+3)*(2*l+1)))
                elif dm == 0:
                    return (-1.)**(m)*np.sqrt(3*(l-m+1)*(l+m+1)/(4*np.pi*(2*l+3)*(2*l+1)))
                elif dm == -1:
                    return (-1.)**(m-1)*np.sqrt(3*(l-m+2)*(l-m+1)/(8*np.pi*(2*l+3)*(2*l+1)))
            elif dl == -1:
                if dm==1:
                    return (-1.)**(m)*np.sqrt(3*(l-m)*(l-m-1)/(8*np.pi*(2*l+1)*(2*l-1)))
                elif dm == 0:
                    return (-1.)**(m)*np.sqrt(3*(l+m)*(l-m)/(4*np.pi*(2*l+1)*(2*l-1)))
                elif dm == -1:
                    return (-1.)**(m)*np.sqrt(3*(l+m)*(l+m-1)/(8*np.pi*(2*l+1)*(2*l-1)))
        else:
            return 0.0
    except ValueError:
        print('Invalid entries for dipole matrix element-related Gaunt coefficients')
        print('l = {:0.4f}, m = {:0.4f}, dl = {:0.4f}, dm = {:0.4f}'.format(l,m,dl,dm))
        return 0.0
    
    
def Yproj(basis):
    '''
    Define the unitary transformation rotating the basis of different inequivalent atoms in the
    basis to the basis of spherical harmonics for sake of defining L.S operator in basis of user
    
    29/09/2018 added reference to the spin character 'sp' to handle rotated systems effectively

    *args:*

        - **basis**: list of orbital objects
    
    *return*:

        - dictionary of matrices for the different atoms and l-shells--keys are tuples of (atom,l)
     
    ***
    '''
    normal_order = {0:{'':0},1:{'x':0,'y':1,'z':2},2:{'xz':0,'yz':1,'xy':2,'ZR':3,'XY':4},3:{'z3':0,'xz2':1,'yz2':2,'xzy':3,'zXY':4,'xXY':5,'yXY':6}}
    a = basis[0].atom
    n = basis[0].n
    l = basis[0].l
    sp = basis[0].spin
    M = {}
    M_tmp = np.zeros((2*l+1,2*l+1),dtype=complex)
    for b in basis:
        if np.linalg.norm(b.Dmat-np.identity(2*b.l+1))>0:
            Dmat = b.Dmat
        else:
            Dmat = None
        loc_rot = b.orient
        label = b.label[2:]

        if b.atom==a and b.n==n and b.l==l and b.spin==sp:
            for p in b.proj:
                M_tmp[l-int(p[-1]),normal_order[l][label]] = p[0]+1.0j*p[1]
                
        else:
                #If we are using a reduced basis, fill in orthonormalized projections for other states in the shell
                #which have been ignored in our basis choice--these will still be relevant to the definition of the LS operator
            M_tmp = fillin(M_tmp,l,Dmat)            
            M[(a,n,l,sp)] = M_tmp
                ##Initialize the next M matrix               
            a = b.atom
            n = b.n
            l = b.l
            sp = b.spin
            M_tmp = np.zeros((2*l+1,2*l+1),dtype=complex)
            for p in b.proj:
                M_tmp[l-int(p[-1]),normal_order[l][label]] = p[0]+1.0j*p[1]
    
    M_tmp = fillin(M_tmp,l,loc_rot)
    M[(a,n,l,sp)] = M_tmp
    
    return M

def fillin(M,l,Dmat=None):
    '''
    If only using a reduced subset of an orbital shell (for example, only t2g states in d-shell),
    need to fill in the rest of the projection matrix with some defaults

    *args*:

        - **M**: numpy array of (2l+1)x(2l+1) complex float

        - **l**: int

        - **Dmat**: numpy array of (2l+1)x(2l+1) complex float

    *return*:

        - **M**: numpy arrayof (2l+1)x(2l+1) complex float

    ***
    '''

    normal_order_rev = {0:{0:''},1:{0:'x',1:'y',2:'z'},2:{0:'xz',1:'yz',2:'xy',3:'ZR',4:'XY'},3:{0:'z3',1:'xz2',2:'yz2',3:'xzy',4:'zXY',5:'xXY',6:'yXY'}}

    for m in range(2*l+1):
        if np.linalg.norm(M[:,m])==0: #if column is empty (i.e. user-defined projection does not exist)
            proj = np.zeros(2*l+1,dtype=complex) 
            for pi in projdict[str(l)+normal_order_rev[l][m]]: 
                proj[l-int(pi[-1])] = pi[0]+1.0j*pi[1] #fill the column with generic projection for this orbital (this will be a dummy)
            if type(Dmat)==np.ndarray:
#                print('l: {:d},'.format(l),'Dmat: ',Dmat,'proj: ',proj)
                proj = np.dot(Dmat,proj)
            for mp in range(2*l+1): #Orthogonalize against the user-defined projections
                if mp!=m:
                    if np.linalg.norm(M[:,mp])!=0:
                        if np.dot(M[:,m],M[:,mp])>1e-10:
                            proj = GramSchmidt(proj,M[:,mp])
            M[:,m] = proj            
    return M
    

def GramSchmidt(a,b):
    '''
    Simple orthogonalization of two vectors, returns orthonormalized vector
    
    *args*:

        - **a**, **b**: numpy array of same length

    *returns*:

        - **GS_a**: numpy array of same size, orthonormalized to the b vector

    ***
    '''
    GS_a = a - np.dot(a,b)/np.dot(b,b)*b
    return GS_a/np.linalg.norm(GS_a)




    

    
if __name__=="__main__":
    x = np.linspace(0,5,100)
    tmp = laguerre(x,5,0)
#    th = np.random.random()*np.pi
#    ph = np.random.random()*2*np.pi
#    for i in range(4):
#        for j in range(-i,i+1):
#            Yme = Y(i,j,th,ph) 
#            Ysc = sc.sph_harm(j,i,ph,th)
#            diff = abs(Yme-Ysc)
#            print i,j,diff
#    


# -*- coding: utf-8 -*-

#Created on Tue Jun 13 17:22:12 2017

#@author: rday

#MIT License

#Copyright (c) 2018 Ryan Patrick Day

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

#This script handles calculation of the Radial Integrals:
#    $int(dr r^3 j_lp(kf*r) R_nl(r))$
#for the photoemission matrix element calculations.
#It loads the radial part of the wavefunction from electron_configs, which itself
#takes Z and orbital label as arguments to do so.
#The integration technique is based on a simple adaptive integrations algorithm
#which is a sort of divide and conquer approach. For each iteration, the interval
#is halved and the Riemann sum is computed. If the difference is below tolerance,
#either one or both of the halves will return. If not, if either or both return diff
#above tolerance, than the corresponding interval is again split and the cycle repeats
#until the total difference over the domain is satisfactorily below tolerance.
#In this way, the partition of the domain is non-regular, with regions of higher
#curvature having a finer partitioning than those with flat regions of the function.


import scipy.special as sc


def general_Bnl_integrand(func,kn,lp):
    
    '''
    Standard form of executable integrand in the e.r approximation of the matrix element
    
    *args*:

        - **func**: executable function of position (float), in units of Angstrom
        
        - **kn**: float, norm of the k vector (in inverse Angstrom)
        
        - **lp**: int, final state angular momentum quantum number
    
    *return*:

        - executable function of float (position)
    
    ***
    '''
    
    def lambda_gen():
        return lambda r: (-1.0j)**lp*r**3*sc.spherical_jn(lp,r*kn)*func(r)
    
    return lambda_gen()

def rect(func,a,b):
    
    '''
    
    Approximation to contribution of a finite domain to the integral, 
    evaluated as a rough rectangle
    
    *args*:

        - **func**: executable to evaluate
        
        - **a**: float, start of interval
        
        - **b**: float, end of interval
        
    *return*: 

        - **recsum**: (complex) float approximated area of the region under
        function between **a** and **b**
    
    ***
    '''
    mid = (a+b)/2.0
    jac = abs(b-a)/6.0
    recsum = jac*(func(a)+4*func(mid)+func(b))
    return recsum

def recursion(func,a,b,tol,currsum):
    '''
    
    Recursive integration algorithm--rect is used to approximate the integral
    under each half of the domain, with the domain further divided until
    result has converged
    
    *args*:

        - **func**: executable
        
        - **a**: float, start of interval
        
        - **b**: float, end of interval
        
        - **tol**: float, tolerance for convergence
        
        - **currsum**: (complex) float, current evaluation for the integral
        
    *return*: 

        - recursive call to the function if not converged, otherwise the result as complex (or real) float
    
    ***
    '''
    mid = (a+b)/2.0
    sum_left = rect(func,a,mid)
    sum_right = rect(func,mid,b)
    err = abs(sum_left+sum_right-currsum)
    if err<=15*tol:
        return sum_left+sum_right
    return recursion(func,a,mid,tol/2.,sum_left)+recursion(func,mid,b,tol/2.,sum_right)

def integrate(func,a,b,tol):
    '''
    Evaluate the integral of **func** over the domain covered by **a**, **b**. This 
    begins by seeding the evaluation with a maximally coarse approximation
    to the integral.
    
    *args*:

        - **func**: executable
        
        - **a**: float, start of interval
        
        - **b**: float, end of interval
        
        - **tol**: float, tolerance for convergence
       
    
    *return*:
    
        - **Q**: (complex) float, value of the integral
    
    ***
    '''

    Q = recursion(func,a,b,tol,rect(func,a,b))
    return Q



# -*- coding: utf-8 -*-
"""
Created on Tue Sep  4 16:27:40 2018

@author: rday
"""

import numpy as np
import matplotlib.pyplot as plt
import chinook.klib as klib

def not_point(point):
    '''
    
    Inverse of point, defined in an N-dimensional binary coordinate frame
    
    *args*:
        
        - **point**: int or numpy array of int between 0 and 1
    
    *return*:
        
        - numpy array of int, NOT gate applied to the binary vector point

    ***
    '''
    return np.mod(point+1,2)

def neighbours(point):
    '''
    
    For an unit cube, we can define the set of 3 nearest neighbours by performing
    the requisite modular sum along one of the three Cartesian axes. In this way,
    for an input point, we can extract its neighbours easily.
    
    *args*:
        
        - **point**: numpy array of 3 int, all either 0 or 1
    
    *return*:
        
        - numpy array of 3x3 int, indicating the neighbours of **point** on the
        unit cube.
        
    ***    
    '''
    tmp = np.array([point,point,point])
    return np.mod(tmp+np.identity(3),2)

def corners():
    '''
    Establish the shortest main diagonal of a cube of points, so as to establish
    the main diagonal for tetrahedral partitioning of the cube
    
    *return*:
        
        **main**: tuple of 2 integers indicating the cube coordinates
        
        **cube**: numpy array of 8 corners (8x3) float
    
    ***
    '''
    cube = np.array([[i,j,k] for k in range(2) for j in range(2) for i in range(2)])
    diagonals = np.array([cube[7-ii]-cube[ii] for ii in range(4)])
    min_ind = np.where(np.linalg.norm(diagonals)==np.linalg.norm(diagonals).min())[0][0]
    main = sorted([min_ind,7-min_ind])
    return main,cube
    
def tetrahedra():
    
    '''
    
    Perform partitioning of a cube into tetrahedra. The indices can then be
    dotted with some basis vector set to put them into the proper coordinate frame.
    
    *return*:
        
        - **tetra**: numpy array of 6 x 4 x 3 int, indicating the corners
        of the 6 tetrahedra
        
    ***
    '''
    main,cube = corners()
    tetra = np.zeros((6,4,3))
    tetra[:,:2,:] = cube[main]
    origin_neighbours = neighbours(main[0])
    for i in range(3):
        origin_neighbours_i = origin_neighbours[i]
        neighbour_options = neighbours(origin_neighbours_i)
        choices = neighbour_options[np.where(np.linalg.norm(neighbour_options-main[0],axis=1)>0)]
        tetra[2*i:2*(i+1),2,:] = np.array([origin_neighbours_i,origin_neighbours_i])
        tetra[2*i:2*(i+1),3,:] = choices
        
    return tetra

def tet_inds():
    '''
    Generate, for a single cube, the tetrahedral designations, 
    for the following conventional numbering:
         6 o ---- o 7       
         /      / |
      4 o ---- o5 o 3
        |      | /
      0 o ---- o 1
      
      with 2 hidden from view (below 6, and behind the line-segment connecting 4-5). 
      Here drawn with x along horizontal, z into plane, y vertical
      Defining the real-index spacing between adjacent cubes in a larger array, we can apply this simple prescription
      to define the spanning tetrahedra over the larger k-mesh
    
    *return*:
        
        - **tetra_inds**: numpy array of integer (6x4), with each
        row containing the index of the 4 tetrahedral vertices. Together, for
        of a set of neighbouring points on a grid, we divide into a set of covering
        tetrahedra which span the volume of the cube.
        
    ***
    '''
#    corn_dict = {'000':0,'100':1,'010':2,'110':3,'001':4,'101':5,'011':6,'111':7}
    corn_dict = {'000':0,'001':1,'100':2,'101':3,'010':4,'011':5,'110':6,'111':7}
    
    tetra_vec = tetrahedra().astype(int)
    tetra_inds = np.array([[corn_dict['{:d}{:d}{:d}'.format(*ti)] for ti in tetra_vec[j]] for j in range(6)])
    return tetra_inds


def gen_mesh(avec,N):
    '''
    Generate a mesh of points in 3-dimensional momentum space over the first
    Brillouin zone. These are defined first in terms of recirocal lattice vectors,
    
    i.e. from 0->1 along each, and then are multiplied by the rec. latt. vectors 
    themselves. Note that this implicitly provides a mesh which is not centred
    at zero, but has an origin at the rec. latt. vector (0,0,0)
    
    *args*:
        
        - **avec**: numpy array of 3x3 float, lattice vectors

        - **N**: int, or tuple of 3 int, indicating the number of points along
        each of the reciprocal lattice vectors        
        
    ***
    '''
    
    if type(N)==int:
        N = (N,N,N)
        
    b_vec = klib.bvectors(avec)
    x,y,z = np.linspace(0,1,N[0]),np.linspace(0,1,N[1]),np.linspace(0,1,N[2])
    X,Y,Z = np.meshgrid(x,y,z)
    X,Y,Z = X.flatten(),Y.flatten(),Z.flatten()
    pts = np.dot(np.array([[X[i],Y[i],Z[i]] for i in range(len(X))]),b_vec)

    return pts

def mesh_tetra(avec,N):
    '''
    An equivalent definition of a spanning grid over the Brillouin zone is just
    one which spans the reciprocal cell unit cell. Translational symmetry imposes
    that this partitioning is equivalent to the conventional definition of the 
    Brillouin zone, with the very big advantage that we can define a rectilinear
    grid which spans this volume in a way which can not be done for most
    Bravais lattices in R3. 
    
    *args*:
        
        - **avec**: numpy array of 3x3 float, lattice vectors
        
        - **N**: int, or iterable of 3 int which define the density of the mesh
        over the Brillouin zone.
        
    *return*:
        
        - **pts**: numpy array of Mx3 float, indicating the points in momentum space
        at the vertices of the mesh
        
        - **mesh_tet**: numpy array of Lx4 int, indicating the L-tetrahedra
        which partition the grid
      
    ***
    '''
    
    if type(N)==int:
        N = (N,N,N)
    pts = gen_mesh(avec,N)

    ti = tet_inds()
    
    mesh_tet = []
    for ii in range(len(pts)):
        test_tetra = propagate(ii,N[2],N[0])[ti]
        for t in test_tetra:
            try:
                if t.max()>=len(pts) or t.max()<0:
                    continue
                else:
                    mesh_tet.append(t)
            except KeyError:
                continue
         
    mesh_tet = np.array(mesh_tet)
    
    if len(mesh_tet)>0:
        return pts,mesh_tet
    else:
        print('WARNING: NO K-POINTS FOUND, CHOOSE A FINER K-POINT GRID')
        return None,None
       

def propagate(i,Nr,Nc):
    
    '''
    Distribute the generic corner numbering convention defined for a cube at the 
    origin to a cube starting at some arbitrary point in our grid. Excludes the
    edge points as starting points, so that all cubes are within the grid.
    
    *args*:
        
        - **i**: int, index of origin
        
        - **Nr**: int, number of rows in grid
        
        - **Nc**: int, number of columns in grid
        
    *return*:
        
        - **numpy array of int, len 8 corresponding to the re-numbering of the
        corners of the cube.
    
    ***
    '''
    if (np.mod(i,Nr)<(Nr-1)) and (np.mod(Nr+i,Nr*Nc)>np.mod(i,Nr*Nc)):
        
        return i + np.array([0,1,Nr,Nr+1,Nr*Nc,Nr*Nc+1,Nr*(1+Nc),Nr*(1+Nc)+1])
    else:
        return -1*np.ones(8)



# -*- coding: utf-8 -*-
"""
Created on Thu Sep 12 14:53:36 2019

@author: rday
"""

#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Mar 24 15:08:00 2019

@author: ryanday
"""


import numpy as np
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.widgets import RadioButtons,Slider
from matplotlib import gridspec
import matplotlib.cm as cm



class interface:
    '''
    This interactive tool is intended for exploring the dataset associated
    with an ARPES simulation using chinook.
    The user can scan through the datacube in each dimension.
    This uses matplotlib natively rather than alternative gui systems in python like Tkinter,
    which makes it a bit more robust across platforms.
    '''    
    
    
    def __init__(self,experiment):
        '''
        Initialize the interface. If necessary, the ARPES intensity is calculated
        in advance. The system is initialized as a cut in constant momentum along
        the y-direction. 
        
        *args*:

            - **experiment**: ARPES experiment object, as defined in *chinook.ARPES_lib*
        
        ***
        '''
        self.experiment = experiment
        self.slice_dict = {'Fixed Kx':1,'Fixed Ky':0,'Fixed Energy':2}
        self.other_dims = {0:(2,1),1:(2,0),2:(1,0)} #relevant axes for rows, columns of each dataset
        try:
            _,self.Imat = self.experiment.spectral()
        except AttributeError:
            print('Matrix Elements have not yet been calculated. Running experiment.datacube() now.')
            self.experiment.datacube()
            print('Matrix element calculation complete.')
            _,self.Imat = self.experiment.spectral()

        self.axes = [np.linspace(*self.experiment.cube[1]),np.linspace(*self.experiment.cube[0]),np.linspace(*self.experiment.cube[2])]
        self.state_coords = np.array([self.axes[0][self.experiment.pks[:,1].astype(int)],self.axes[1][self.experiment.pks[:,2].astype(int)],self.experiment.pks[:,3]]) ## Y, X , Energy coordinates of states
        self.state_coords[2,:] = self.bin_energy() #energy is discretized according to the energy mesh values to allow for state selection from angular maps

        self.aspects = [(self.experiment.cube[self.other_dims[ii][0]][1]-self.experiment.cube[self.other_dims[ii][0]][0])/(self.experiment.cube[self.other_dims[ii][1]][1]-self.experiment.cube[self.other_dims[ii][1]][0]) for ii in range(3)]        
        self.extents = [[self.axes[self.other_dims[ii][0]][0],self.axes[self.other_dims[ii][0]][-1],self.axes[self.other_dims[ii][1]][0],self.axes[self.other_dims[ii][1]][-1]] for ii in range(3)]
        self.deltas = [self.axes[0][1]-self.axes[0][0],self.axes[1][1]-self.axes[1][0],self.axes[2][1]-self.axes[2][0]]
        
        self.run_gui()
    
    def run_gui(self):

        '''
        Execution of the matplotlib gui. The figure is initialized, along with all widgets and 
        chosen datasets. The user has access to both the slice of ARPES data plotted, in addition
        to the orbital projection plotted in upper right panel.
        
        '''
        
        self.fig = plt.figure(figsize=(11,11))
        self.fig.canvas.set_window_title('Chinook Orbital Mapper')
        self.ax1 = plt.subplot2grid((10,10),(0,0),rowspan=6,colspan=6,fig=self.fig) #main figure
        self.ax2 = plt.subplot2grid((10,10),(0,7),rowspan=6,colspan=3,fig=self.fig) #subfigure, right
        self.ax3 = plt.subplot2grid((10,10),(7,0),rowspan=3,colspan=6,fig=self.fig) #subfigure, bottom
        self.ax4 = plt.subplot2grid((10,10),(7,7),rowspan=1,colspan=3,fig=self.fig) #slider axis
        self.ax5 = plt.subplot2grid((10,10),(8,7),rowspan=3,colspan=3,fig=self.fig) #radio button axis

        self.dim = 0 #initialize looking at constant Y contour, midway through dataset
        self.indx = int(len(self.axes[0])/2)

        self.img = self.ax1.imshow(self.Imat[self.indx,:,:],cmap=cm.magma,extent=self.extents[self.dim],origin='lower')
        
        self.ax1.set_aspect(self.aspects[self.dim])
        
        self.slide_w = Slider(self.ax4,'',self.axes[0][0],self.axes[0][-1],valinit=self.axes[0][self.indx],valstep =self.deltas[self.dim],color='#1F1B33')
        
        self.radio_ax = RadioButtons(self.ax5,('Fixed Kx','Fixed Ky','Fixed Energy'),active=1,activecolor='#1F1B33')
        
        
        ## INITIALIZE THE ORBITAL PLOT ##
        
        ind1,ind2 = self.other_dims[self.dim]  ##rows, columns
        self.cursor_location =  np.array([self.axes[ind1][int(len(self.axes[ind1])/2)],self.axes[ind2][int(len(self.axes[ind2])/2)]]) ##row, column coordinates of centre, in phys. units
        self.rowdc, = self.ax2.plot(self.Imat[self.indx,:,int(len(self.axes[ind1])/2)],self.axes[ind2]) #all rows distribution curve, here fixed energy, all kx
        self.coldc, = self.ax3.plot(self.axes[ind1],self.Imat[self.indx,int(len(self.axes[ind2])/2),:]) #all column distribution curve, here fixed kx, all energy
        self.cursor, = self.ax1.plot([self.cursor_location[0]],[self.cursor_location[1]],marker='+',markersize=20,c='r')
        self.ax1.set_title('Cursor: E: {:0.04f}, X: {:0.04f}'.format(self.cursor_location[0],self.cursor_location[1]))

       

        self.plot_peaks = self.state_coords[(ind1,ind2),:][:,np.where(self.state_coords[self.dim,:]==self.axes[self.dim][self.indx])[0]]
        self.bands, = self.ax1.plot(self.plot_peaks[0,:],self.plot_peaks[1,:],marker='o',linestyle='',markersize=2,c='w',alpha=0.3)
        
        
        
        ## SLIDER FUNCTION ##
        def img_slide(val):

            '''
            User requests another slice of the image to be displayed using the 
            slider widget. Image is updated with the requested dataset.
            
            *args*:

                - **val**: float, slider value chosen

            ***
            '''
            self.indx = int((val-self.axes[self.dim][0])/(self.axes[self.dim][1]-self.axes[self.dim][0]))
            self.plot_img()
            self.find_cursor()
            self.fig.canvas.draw()
            plt.show()
            
            
        ## SLICE SELECTION RADIO BUTTON FUNCTION ##
        def button_click(label):
            '''

            User requests plotting of intensity map along an alternative axis.
            The intensity map is regenerated with axes scaled to reflect the 
            extent of data within the new plane.
            
            *args*:

                - **label**: string, radio-button label associated with the 
                possible slice directions

            ***    
            '''
            self.dim = self.slice_dict[label]  
            
            ind1,ind2 = self.other_dims[self.dim]
            self.cursor_location = np.array([self.axes[ind1][int(len(self.axes[ind1])/2)],self.axes[ind2][int(len(self.axes[ind2])/2)]]) #column, row in physical units
            
            self.indx = int(len(self.axes[self.dim])/2)
            if self.dim==0:
                
                self.img = self.ax1.imshow(self.Imat[self.indx,:,:],cmap=cm.magma,extent=self.extents[self.dim],origin='lower')
                self.img.set_clim(vmin = self.Imat[self.indx,:,:].min()*0.8,vmax=self.Imat[self.indx,:,:].max()*1.2)

            elif self.dim==1:
                self.img = self.ax1.imshow(self.Imat[:,self.indx,:],cmap=cm.magma,extent=self.extents[self.dim],origin='lower')
                self.img.set_clim(vmin = self.Imat[:,self.indx,:].min()*0.8,vmax=self.Imat[:,self.indx,:].max()*1.2)

            elif self.dim==2:
                self.img = self.ax1.imshow(self.Imat[:,:,self.indx],cmap=cm.magma,extent=self.extents[self.dim],origin='lower')
                self.img.set_clim(vmin = self.Imat[:,:,self.indx].min()*0.8,vmax=self.Imat[:,:,self.indx].max()*1.2)
            
            self.find_cursor()

            self.plot_peaks = self.state_coords[(ind1,ind2),:][:,np.where(self.state_coords[self.dim,:]==self.axes[self.dim][self.indx])[0]]
            self.bands.set_data(self.plot_peaks[0,:],self.plot_peaks[1,:])

            self.slide_w.valmin = self.axes[self.dim][0]
            self.slide_w.valmax = self.axes[self.dim][-1]
            self.slide_w.ax.set_xlim(self.slide_w.valmin,self.slide_w.valmax)
            self.slide_w.delta = self.deltas[self.dim]
            self.slide_w.val = self.axes[self.dim][self.indx]
            
            self.ax1.set_aspect(self.aspects[self.dim])
  
            self.fig.canvas.draw()
        
        def onclick(event):

            '''
            User selects a point of interest in the intensity map, and the 
            corresponding orbital eigenvector is displayed in the upper right 
            panel of the window.
            
            *args*:

                - **event**: button_press_event associated with mouse query 
                within the frame of self.ax1. All other mouse events are
                disregarded.

            ***
            '''
            if event.inaxes==self.ax1.axes:
                coords = np.array([event.xdata,event.ydata])
                ind1,ind2 = self.other_dims[self.dim]
                self.cursor_location = np.array([coords[0],coords[1]])
                self.find_cursor()

                self.fig.canvas.draw()
                plt.show()
                
                
        cid = self.fig.canvas.mpl_connect('button_press_event',onclick)
        self.slide_w.on_changed(img_slide)
        self.radio_ax.on_clicked(button_click)
        plt.tight_layout()
        plt.show()
        
    def find_cursor(self):

        '''
        Find nearest point to the desired cursor position, as clicked by the
        user. The cursor event coordinates are compared against the peak positions
        from the tight-binding calculation, and a best choice within the plotted 
        slice is selected.
        
        *args*:

            - **cursor**: tuple of 2 float, indicating the column and row of the
            event, in units of the data-set scaling (e.g. 1/A or eV)
        
        ***
        '''
        ind1,ind2 = self.other_dims[self.dim] #columns,rows
        cursor = self.cursor_location
        
        axis1 = self.axes[ind1] #
        axis2 = self.axes[ind2]

        row_index = np.where((abs(axis1-cursor[0])==abs(axis1-cursor[0]).min()))[0][0]
        col_index = np.where((abs(axis2-cursor[1])==abs(axis2-cursor[1]).min()))[0][0]
        
        if self.dim == 0:
            rowdc = self.Imat[self.indx,:,row_index]
            coldc = self.Imat[self.indx,col_index,:]
            imax = self.Imat[self.indx,:,:].max()
            self.ax1.set_title('Cursor: E: {:0.04f}, Kx: {:0.04f}'.format(cursor[0],cursor[1]))

        if self.dim == 1:
            rowdc = self.Imat[:,self.indx,row_index]
            coldc = self.Imat[col_index,self.indx,:]
            imax = self.Imat[:,self.indx,:].max()
            self.ax1.set_title('Cursor: E: {:0.04f}, Ky: {:0.04f}'.format(cursor[0],cursor[1]))

        if self.dim == 2:
            rowdc = self.Imat[:,row_index,self.indx]
            coldc = self.Imat[col_index,:,self.indx]
            imax = self.Imat[:,:,self.indx].max()
            self.ax1.set_title('Cursor: Kx: {:0.04f}, Ky: {:0.04f}'.format(cursor[0],cursor[1]))

        
        self.rowdc.set_data(rowdc,axis2)
        self.coldc.set_data(axis1,coldc)
        self.cursor.set_data([cursor[0]],[cursor[1]])
        self.ax2.set_ylim(axis2.min(),axis2.max())
        self.ax2.set_xlim(rowdc.min(),imax)
        self.ax3.set_ylim(coldc.min(),imax)
        self.ax3.set_xlim(axis1.min(),axis1.max())    

        
        
    def plot_img(self):

        '''
        Update the plotted intensity map slice. The plotted bandstructure states are 
        also displayed.
        
        ***
        '''
        ind1,ind2 = self.other_dims[self.dim]
        self.plot_peaks = self.state_coords[(ind1,ind2),:][:,np.where(self.state_coords[self.dim,:]==self.axes[self.dim][self.indx])[0]]
        self.bands.set_data(self.plot_peaks[0,:],self.plot_peaks[1,:])
        if self.dim ==0:
            self.img.set_data(self.Imat[self.indx,:,:])
            self.img.set_clim(vmin = self.Imat[self.indx,:,:].min()*0.8,vmax=self.Imat[self.indx,:,:].max()*1.2)

        elif self.dim == 1:
            self.img.set_data(self.Imat[:,self.indx,:])
            self.img.set_clim(vmin = self.Imat[:,self.indx,:].min()*0.8,vmax=self.Imat[:,self.indx,:].max()*1.2)

        elif self.dim == 2:
            self.img.set_data(self.Imat[:,:,self.indx])
            self.img.set_clim(vmin = self.Imat[:,:,self.indx].min()*0.8,vmax=self.Imat[:,:,self.indx].max()*1.2)
            
    def bin_energy(self):

        '''
        Translate the exact energy value for the band peaks into the discrete
        binning of the intensity map, to allow for cursor queries to be processed.
        
        *return*:

            - **coarse_pts**: numpy array of float, same lengths as *self.state_coords*,
            but sampled over a discrete grid.

        ***
        '''
        
        fine_pts = self.state_coords[2,:]
        coarse_pts = np.array([self.axes[2][int((fine_pts[ii]-self.axes[2][0])/(self.axes[2][1]-self.axes[2][0]))] if (fine_pts[ii]<self.axes[2].max() and fine_pts[ii]>self.axes[2].min()) else -1 for ii in range(len(fine_pts))])
        return coarse_pts


if __name__ == "__main__":
    plotimg = interface(expmt)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Dec 28 13:33:59 2018

@author: ryanday
"""

import numpy as np
from collections import Iterable
import matplotlib.pyplot as plt


hb = 6.626e-34/(2*np.pi)
kb = 1.38e-23
q = 1.602e-19
me = 9.11e-31
A = 1e-10


def k_parallel(ek):
    '''
    Convert kinetic energy in eV to inverse Angstrom
    '''
    return np.sqrt(2*me/hb**2*ek*q)*A


def ang_mesh(N,th,ph):
    '''
    Generate a mesh over the indicated range of theta and phi,
    with N elements along each of the two directions
    
    *args*:

        - **N**: int or iterable of length 2 indicating # of points along *th*, and *ph* respectively
        
        - **th**: iterable length 2 of float (endpoints of theta range)
        
        - **ph**: iterable length 2 of float (endpoints of phi range)
    
    *return*:

        - numpy array of N_th x N_ph float, representing mesh of angular coordinates
    
    ***
    '''
    try:
        if isinstance(N,Iterable):
            N_th,N_ph = int(N[0]),int(N[1])
        else:
            N_th,N_ph = int(N),int(N)
    except ValueError:
        print('ERROR: Invalid datatype for number of points in mesh. Return none')
        return None
    return np.meshgrid(np.linspace(th[0],th[1],N_th),np.linspace(ph[0],ph[1],N_ph))

def k_mesh(Tmesh,Pmesh,ek):
    '''
    Application of rotation to a normal-emission vector (i.e. (0,0,1) vector)
    Third column of a rotation matrix formed by product of rotation about vertical (ky), and rotation around kx axis
    c.f. Labbook 28 December, 2018
    
    *args*:

        - **Tmesh**: numpy array of float, output of *ang_mesh*
        
        - **Pmesh**: numpy array of float, output of *ang_mesh*
        
        - **ek**: float, kinetic energy in eV
    
    *return*:

        - **kvec**: numpy array of float, in-plane momentum array associated with angular emission coordinates
    
    ***
    '''
    klen = k_parallel(ek)
    kvec = klen*np.array([-np.sin(Tmesh),-np.cos(Tmesh)*np.sin(Pmesh),np.cos(Tmesh)*np.cos(Pmesh)])

    return kvec



def rot_vector(vector,th,ph):
    '''
    Rotation of vector by theta and phi angles, about the global y-axis by theta, followed by a rotation about
    the LOCAL x axis by phi. This is analogous to the rotation of a cryostat with a vertical-rotation axis (theta),
    and a sample-mount tilt angle (phi). NOTE: need to extend to include cryostats where the horizontal rotation axis
    is fixed, as opposed to the vertical axis--I have never seen such a system but it is of course totally possible.
    
    *args*:

        - **vector**: numpy array length 3 of float (vector to rotate)
        
        - **th**: float, or numpy array of float -- vertical rotation angle(s)
        
        - **ph**: float, or numpy array of float -- horizontal tilt angle(s)
    
    *return*:

        - numpy array of float, rotated vectors for all angles: shape 3 x len(ph) x len(th)
        NOTE: will flatten any length-one dimensions  
    
    ***
    '''
   
    if not isinstance(th,Iterable):
        th = np.array([th])
    if not isinstance(ph,Iterable):     
        ph = np.array([ph])
    th,ph = np.meshgrid(th,ph)
     
    Rm = np.array([[np.cos(th),np.zeros(np.shape(th)),np.sin(th)],[np.sin(th)*np.sin(ph),np.cos(ph),-np.cos(th)*np.sin(ph)],[-np.sin(th)*np.cos(ph),np.sin(ph),np.cos(th)*np.cos(ph)]])
    return np.squeeze(np.einsum('ijkl,j->ikl',Rm,vector))


def plot_mesh(ek,N,th,ph):
    '''
    Plotting tool, plot all points in mesh, for an array of N angles,
    at a fixed kinetic energy.

    *args*:

        - **ek**: float, kinetic energy, eV
        
        - **N**: tuple of 2 int, number of points along each axis
        
        - **thx**: tuple of 2 float, range of horizontal angles, radian
        
        - **thy**: tuple of 2 float, range of vertical angles, radian

    ***
    '''

    kpts = gen_kpoints(ek,N,th,ph,0)
    
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.scatter(kpts[:,0],kpts[:,1])
    ax.set_aspect(1)


def gen_kpoints(ek,N,thx,thy,kz):
    '''
    Generate a mesh of kpoints over a mesh of emission angles.
    
    *args*:

        - **ek**: float, kinetic energy, eV
        
        - **N**: tuple of 2 int, number of points along each axis
        
        - **thx**: tuple of 2 float, range of horizontal angles, radian
        
        - **thy**: tuple of 2 float, range of vertical angles, radian
        
        - **kz**: float, k-perpendicular of interest, inverse Angstrom
        
    *return*:

        - **k_array**: numpy array of N[1]xN[0] float, corresponding to mesh of in-plane momenta
        
    ***    
    '''
    Thx,Thy = ang_mesh(N,thx,thy)
    kv = k_mesh(Thx,Thy,ek)
    kx = kv[0,:,:].flatten()
    ky = kv[1,:,:].flatten()
    k_array = np.array([[kx[i],ky[i],kz] for i in range(len(kx))])
    return k_array


if __name__ == "__main__":
    rad = np.pi/180
    kp = k_parallel(60)
    N = 37
    To = 0
    Tlims = [-13*rad+To*rad,13*rad+To*rad]
    Plims = [-25*rad,25*rad]
    #plot_mesh(10,N,Tlims,Plims)
    
    kp = gen_kpoints(21,(20,30),(-1,0),(0,3),0)
    fig  = plt.figure()
    ax = fig.add_subplot(111)
    ax.scatter(kp[:,0],kp[:,1])


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#Created on Thu Feb 22 11:34:32 2018

#@author: rday

#MIT License

#Copyright (c) 2018 Ryan Patrick Day

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

import matplotlib
from matplotlib import rc
from matplotlib import rcParams
import os
rcParams.update({'figure.autolayout':True})


import numpy as np

from matplotlib.figure import Figure
import matplotlib.cm as cm

try:
    import tkinter as Tk
    from tkinter import messagebox
    from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
    tk_found = True
except ModuleNotFoundError:
    print('tkinter not found, please load for interactive map plotting')
    tk_found = False



import chinook.intensity_map as imap


    
    
def tk_query():
    if tk_found:
        return True
    else:
        return False

    
if tk_found:

    class plot_intensity_interface:
        '''
        Tkinter-based plotting interface. Interactive, allows for generation
        and exploration of intensity maps, given an existing TB and experiment
        object.
        '''
        
        def __init__(self,experiment):
            self.root = Tk.Tk()
            self.experiment = experiment
            print('Initializing spectral function...')
            try:
                _,self.Imat = self.experiment.spectral()
            except AttributeError:
                print('Matrix elements have not yet been computed. Initializing matrix elements now...')
                self.experiment.datacube()
                print('Matrix element calculation complete.')
                _,self.Imat = self.experiment.spectral()
                
            map_pars = (0,self.Imat,self.experiment.cube,self.experiment.kz,self.experiment.T,self.experiment.hv,self.experiment.pol,self.experiment.dE,self.experiment.dk,self.experiment.SE_args,self.experiment.sarpes,self.experiment.ang)
    
            self.Imat_dict = {'I_0':imap.intensity_map(*map_pars)}
            self.x = self.experiment.cube[0]
            self.y = self.experiment.cube[1]
            self.w = self.experiment.cube[2]
            self.dx,self.dy,self.dw = (self.x[1]-self.x[0])/self.x[2],(self.y[1]-self.y[0])/self.y[2],(self.w[1]-self.w[0])/self.w[2]
            print('Initializing interface...')
            self.plot_make()
    
        def plot_make(self):
               
            self.root.wm_title('CHINOOK DATA PLOTTER')
    
            fig1 = Figure(figsize=(5,5))
            fig2 = Figure(figsize=(5,5))
            fig3 = Figure(figsize=(5,5))
            
            rc('font',**{'family':'serif','serif':['Palatino'],'size':12})
            rc('text',usetex = False) 
    
            ax1= fig1.add_subplot(111)
            ax2= fig2.add_subplot(111)
            ax3= fig3.add_subplot(111)
            sp1 = ax1.imshow(self.Imat[:,:,int(self.w[2]/2)],extent=[self.x[0],self.x[1],self.y[1],self.y[0]],cmap=cm.magma) ##IMSHOW HAS INVERTED ORIGIN CONVENTION TO PCOLORMESH--ORIGIN IS AT TOP-LEFT INSTEAD OF BOTTOM-LEFT
            sp2 = ax2.imshow(self.Imat[:,int(self.x[2]/2),:],extent=[self.w[0],self.w[1],self.y[1],self.y[0]],cmap=cm.magma)
            sp3 = ax3.imshow(self.Imat[int(self.y[2]/2),:,:],extent=[self.w[0],self.w[1],self.x[1],self.x[0]],cmap=cm.magma)
            sp1.set_clim(vmin=0,vmax=self.Imat.max()*1.05)
            sp2.set_clim(vmin=0,vmax=self.Imat.max()*1.05)
            sp3.set_clim(vmin=0,vmax=self.Imat.max()*1.05)
            try:
                asp1 = (self.x[1]-self.x[0])/(self.y[1]-self.y[0])
            except ZeroDivisionError:
                asp1=1.0
            try:
                asp2 = (self.w[1]-self.w[0])/(self.y[1]-self.y[0])
            except ZeroDivisionError:
                asp2=1.0
            try:
                asp3 = (self.w[1]-self.w[0])/(self.x[1]-self.x[0])
            except ZeroDivisionError:
                asp3 =1.0
            ax1.set_aspect(asp1)
            ax2.set_aspect(asp2)
            ax3.set_aspect(asp3)
            
            ax1.tick_params(axis='both',labelsize=12)
            ax2.tick_params(axis='both',labelsize=12)
            ax3.tick_params(axis='both',labelsize=12)
            
            ax1.set_ylabel('Momentum y ($\AA^{-1}$)')
            ax1.set_xlabel('Momentum x ($\AA^{-1}$)')
            ax2.set_ylabel('Momentum y ($\AA^{-1}$)')
            ax2.set_xlabel('Energy (eV)')
            ax3.set_ylabel('Momentum x ($\AA^{-1}$)')
            ax3.set_xlabel('Energy (eV)')
    
            plt_win_1 = FigureCanvasTkAgg(fig1,master=self.root)
            plt_win_1.show()
            plt_win_1.get_tk_widget().grid(row=0,column=0,columnspan=3,rowspan=3)
            
            plt_win_2 = FigureCanvasTkAgg(fig2,master=self.root)
            plt_win_2.show()
            plt_win_2.get_tk_widget().grid(row=0,column=3,columnspan=3,rowspan=3)
    
            plt_win_3 = FigureCanvasTkAgg(fig3,master=self.root)
            plt_win_3.show()
            plt_win_3.get_tk_widget().grid(row=0,column=6,columnspan=3,rowspan=3)
    
    
            def _quit_win():
                self.root.quit()
                self.root.destroy()
        
            def _update_slice(event):
                sv1 = int((float(slide_1.get())-self.w[0])/(self.dw))
                sv2 = int((float(slide_2.get())-self.x[0])/(self.dx))
                sv3 = int((float(slide_3.get())-self.y[0])/(self.dy))
        
                sp1.set_data(self.Imat[:,:,sv1])
                sp2.set_data(self.Imat[:,sv2,:])
                sp3.set_data(self.Imat[sv3,:,:])
                fig1.canvas.draw()
                fig2.canvas.draw()
                fig3.canvas.draw()
        
            s1_label = Tk.Label(master=self.root,text="Energy (eV)").grid(row=3,column=0)
            slide_1 = Tk.Scale(master=self.root,from_=self.w[0],to=(self.w[1]-self.dw),orient='horizontal',resolution=self.dw,command=_update_slice)
            slide_1.grid(row=3,column=1,sticky='W')
    
            s2_label = Tk.Label(master=self.root,text="Momentum Kx (1/Å)").grid(row=3,column=3)
            slide_2 = Tk.Scale(master=self.root,from_=self.x[0],to=(self.x[1]-self.dx),orient='horizontal',resolution=self.dx,command=_update_slice)
            slide_2.grid(row=3,column=4,sticky='W')
    
            s3_label = Tk.Label(master=self.root,text="Momentum K_y (1/Å)").grid(row=3,column=6)
            slide_3 = Tk.Scale(master=self.root,from_=self.y[0],to=(self.y[1]-self.dy),orient='horizontal',resolution=self.dy,command=_update_slice)
            slide_3.grid(row=3,column=7,sticky='W')
            
            
            def _customize():
                
                cwin = Tk.Toplevel()
                cwin.wm_title('CUSTOMIZE MAP DISPLAY')
                #List of available Maps
                list_lab = Tk.Label(master=cwin,text='Available maps: ').grid(row=0,column=0)
                mat_list = (d for d in self.Imat_dict)
                mat_listbox = Tk.Listbox(master=cwin)
                mat_listbox.grid(row=1,column=0,columnspan=4)
                for d in list(enumerate(self.Imat_dict)):
                    mat_listbox.insert("end",d[1])
                #Spin Vector
                spin_label=Tk.Label(master=cwin,text='Spin Axis: ').grid(row=2,column=0)
                sx = Tk.Entry(master=cwin)
                sy = Tk.Entry(master=cwin)
                sz = Tk.Entry(master=cwin)
                
                sx.insert('end','0.0')
                sy.insert('end','0.0')
                sz.insert('end','1.0')
                
                sx.grid(row=2,column=1)
                sy.grid(row=2,column=2)
                sz.grid(row=2,column=3)
                
                #Spin projection -- -1, None, 1
                proj_label = Tk.Label(master=cwin,text='Spin Projection: ').grid(row=3,column=0)
                proj_list = ("Down","Up","None")
                proj_var = Tk.StringVar()
                proj_var.set(proj_list[2])
                proj_opt = Tk.OptionMenu(cwin,proj_var,*proj_list)
                proj_opt.grid(row=3,column=1)
                #Polarization Vector
                pol_label = Tk.Label(master=cwin,text='Polarization: ').grid(row=4,column=0)
                pol_x = Tk.Entry(master=cwin)
                pol_y = Tk.Entry(master=cwin)
                pol_z = Tk.Entry(master=cwin)
                
                pol_x.insert('end','1.0+0.0j')
                pol_y.insert('end','0.0+0.0j')
                pol_z.insert('end','0.0+0.0j')
                
                pol_x.grid(row=4,column=1)
                pol_y.grid(row=4,column=2)
                pol_z.grid(row=4,column=3)
                #Name of New Map -- option of string or Blank
                
                ###SELF ENERGY
                def _self_energy():
                    
                    sewin = Tk.Toplevel()
                    sewin.wm_title('SELF ENERGY')
                    
                    SE_label = Tk.Label(master=sewin,text='Im(Self-Energy)').grid(row=0,column=0)
                    SE_entry = Tk.Entry(master=sewin)
                    SE_entry.grid(row=0,column=1)
    
                    def _se_infobox():
                        messagebox.showinfo('Information','The imaginary part of the self energy can be entered here. The absolute value of the function will be taken, preserving particle-hole symmetry and causality.',parent=sewin)
                
                    def _quit_sub():
                        sewin.quit()
                        sewin.destroy()
                    
                    se_info_button = Tk.Button(master=sewin,text='Info',command=_se_infobox)
                    se_info_button.grid(row=1,column=0,columnspan=1)        
                    
                    SEquit = Tk.Button(master=sewin,text='Quit',command=_quit_sub)
                    SEquit.grid(row=1,column=1)
                    
                    sewin.mainloop()
                
                SE_button = Tk.Button(master=cwin,text='Self-Energy',command=_self_energy)
                SE_button.grid(row=5,column=0)
                
                mat_nm_label = Tk.Label(master=cwin,text='Map Name: ').grid(row=6,column=0)
                mat_nm = Tk.Entry(master=cwin)
                mat_nm.grid(row=6,column=1)
            
                
                #Add a new map to the list of available maps
                def _add_map():
                    
                    tmp_dict = {}
                    #Define the polarization vector
                    tmp_dict['pol'] = xyz(pol_x.get(),pol_y.get(),pol_z.get())
                    
                    proj_choice = proj_var.get()
                    if proj_choice=="None":
                        tmp_dict['spin']=None
                    else:
                        
                        tmp_dict['spin'] = [-1 if proj_choice=="Down" else 1,np.array([float(sx.get()),float(sy.get()),float(sz.get())])]
                    mat_name = mat_nm.get() if mat_nm.get()!="" else "I_{:d}".format(len(self.Imat_dict))
    
                    map_pars = (self.experiment.cube,self.experiment.kz,self.experiment.T,self.experiment.hv,tmp_dict['pol'],self.experiment.dE,self.experiment.dk,self.experiment.SE_args,tmp_dict['spin'],self.experiment.ang)
                    _,Imap = self.experiment.spectral(ARPES_dict = tmp_dict)
                    
                    self.Imat_dict[mat_name] = imap.intensity_map(len(self.Imat_dict.keys()),Imap,*map_pars)
    #                self.meta[mat_name] = tmp_dict #Save the parameters associated with a given calculation for use in export
                    
                    mat_listbox.insert("end",mat_name)            
                
                add_button = Tk.Button(master=cwin,text ='Add Map ',command=_add_map)
                add_button.grid(row=6,column=2,columnspan=1)            
                
                #add option of operating on datasets
                def _gen_map():
                    st_raw = op_entry.get()
                    placeholder  = ''
                    for d in self.Imat_dict:
                        before = st_raw
                        replacement = 'self.Imat_dict["{:s}"].Imat'.format(d)
                        st_raw = st_raw.replace(d,replacement)
                        if st_raw!=before:
                            placeholder = d
                        self.Imat_dict[d].Imat+=  abs(self.Imat_dict[d].Imat[np.nonzero(self.Imat_dict[d].Imat)]).min()*10**-4#avoid divergence for division if zeros present
                    st_raw = st_raw.replace("SQRT","np.sqrt")
                    st_raw = st_raw.replace("COS","np.cos")
                    st_raw = st_raw.replace("SIN","np.sin")
                    st_raw = st_raw.replace("TAN","np.tan")
                    st_raw = st_raw.replace("EXP","np.exp")
                    st_raw = st_raw.replace("LOG","np.log")
                    st_raw = st_raw.replace("^","**")
                    
                    tmp_mat = eval(st_raw)
                    map_nm = mat_nm.get() if (mat_nm.get()!="" or bool(sum([mat_nm==d for d in self.Imat_dict]))) else "I_{:d}".format(len(self.Imat_dict))
                    mat_listbox.insert("end",map_nm)
    #                tmp_mat = eval(replacement).copy()
                    self.Imat_dict[map_nm] = self.Imat_dict[placeholder].copy()
                    self.Imat_dict[map_nm].Imat = tmp_mat.copy()
                    self.Imat_dict[map_nm].notes = 'Intensity calculated as: {:s}'.format(st_raw)
    
                    
            
                op_label = Tk.Label(master=cwin,text="Operate: ").grid(row=7,column=0)
                op_entry = Tk.Entry(master=cwin)
                op_entry.grid(row=7,column=1)
                op_gen = Tk.Button(master=cwin,text = "Generate",command=_gen_map)
                op_gen.grid(row=7,column=2)
                
                def _plot_map():
                    selected = mat_listbox.curselection()[0]
                    try:
                        
                        self.Imat = self.Imat_dict[mat_listbox.get(selected)].Imat
                        sv1 = int((float(slide_1.get())-self.w[0])/(self.dw))
                        sv2 = int((float(slide_2.get())-self.x[0])/(self.dx))
                        sv3 = int((float(slide_3.get())-self.y[0])/(self.dy))
        
                        sp1.set_data(self.Imat[:,:,sv1])
                        sp2.set_data(self.Imat[:,sv2,:])
                        sp3.set_data(self.Imat[sv3,:,:])
                        sp1.set_clim(vmin=self.Imat.min(),vmax=self.Imat.max())
                        sp2.set_clim(vmin=self.Imat.min(),vmax=self.Imat.max())
                        sp3.set_clim(vmin=self.Imat.min(),vmax=self.Imat.max())
                        fig1.canvas.draw()
                        fig2.canvas.draw()
                        fig3.canvas.draw()
                        cwin.quit()
                        cwin.destroy()
                        
                    except IndexError:
                        print('Please select a map to plot before attempting to plot.')
                
                plot_button = Tk.Button(master=cwin,text="Plot Selection", command=_plot_map)
                plot_button.grid(row=8,column=0)
                
                def _infobox():
                    messagebox.showinfo("Information","This panel allows to generate intensity maps with or without\n spin projection, for arbitrary polarization. You can also enter a\n function over several intensity maps, referencing them by names indicated in the list of available maps. Standard polynomial\n operations as well as SQRT, COS, SIN, TAN, EXP, and LOG functions\n can be passed.",parent=cwin) 
                
                info_button = Tk.Button(master=cwin,text='Info',command=_infobox)
                info_button.grid(row=8,column=2)
                
                def _quit_sub():
                    cwin.quit()
                    cwin.destroy()
                
                close_button = Tk.Button(master= cwin,text='Close Window',command=_quit_sub)
                close_button.grid(row=8,column=3,columnspan=1)
                cwin.mainloop()
                
                
            
            custmap_button = Tk.Button(master=self.root,text = 'Custom Map', command = _customize)
            custmap_button.grid(row=5,column=0)
            
            clabel = Tk.Label(master=self.root,text='Colourmap: ').grid(row=4,column=0)
            cmaps = ('Spectral','Blue-White-Red','Magma','Inferno','Black-to-White','White-to-Black')
            cmdict = {'Spectral':cm.Spectral,'Blue-White-Red':cm.bwr,'Magma':cm.magma,'Inferno':cm.inferno,'Black-to-White':cm.Greys_r,'White-to-Black':cm.Greys}
            cmap_var = Tk.StringVar()
            cmap_var.set(cmaps[2])
            cmap_opt = Tk.OptionMenu(self.root,cmap_var,*cmaps)
            cmap_opt.grid(row=4,column=1)
                
            vmin_label = Tk.Label(master=self.root,text='Scale Min: ').grid(row=4,column=2)
            vmin_ent = Tk.Entry(master=self.root)
            vmin_ent.grid(row=4,column=3)
            vmin_ent.insert('end','Auto')
            vmax_label = Tk.Label(master=self.root,text='Scale Max: ').grid(row=4,column=4)
            vmax_ent = Tk.Entry(master=self.root)
            vmax_ent.grid(row=4,column=5)
            vmax_ent.insert('end','Auto')
                
    
            def _update():
                doit = True
                try:
                    vmin = float(vmin_ent.get())
                except ValueError:
                    if vmin_ent.get().upper()=='AUTO':
                        vmin = self.Imat.min()
                    else:
                        print( 'Please enter a numeric value for Scale Minimum ')
                        doit = False
                try:
                    vmax = float(vmax_ent.get())
                except ValueError:
                    if vmax_ent.get().upper()=='AUTO':
                        vmax = self.Imat.max()
                    else:
                        print ('Please enter a numeric value for Scale Maximum ')
                        doit = False
                if vmin>=vmax:
                    doit= False
                if doit:
                    sp1.set_clim(vmin=vmin,vmax=vmax)
                    sp2.set_clim(vmin=vmin,vmax=vmax)
                    sp3.set_clim(vmin=vmin,vmax=vmax)
                    sp1.set_cmap(cmdict[cmap_var.get()])
                    sp2.set_cmap(cmdict[cmap_var.get()])
                    sp3.set_cmap(cmdict[cmap_var.get()])
                    fig1.canvas.draw()
                    fig2.canvas.draw()
                    fig3.canvas.draw()
                        
            col_up = Tk.Button(master=self.root,text='Update Colourscale',command=_update)
            col_up.grid(row=4,column=6)
            
            def _infobox_main():
                messagebox.showinfo("Information","This plotting GUI allows for the user to scan through\n different slices of the momentum-energy cube generated\n in an ARPES experiment. The 'Custom Map' panel allows you to\n change the polarization and spin projection (for Spin-ARPES) as well as \n to produce composite maps generated as a function of maps with different experimental parameters.",parent=self.root) 
                
            main_info_button = Tk.Button(master=self.root,text='Info',command=_infobox_main)
            main_info_button.grid(row=5,column=1)
            
            def _exp_win():
                
                ewin = Tk.Toplevel()
                ewin.wm_title('EXPORT MAPS TO FILE')
    #            List of available Maps
                list_lab = Tk.Label(master=ewin,text='Available maps: ').grid(row=0,column=0)
                mat_list = (d for d in self.Imat_dict)
                mat_listbox = Tk.Listbox(master=ewin)
                mat_listbox.grid(row=1,column=0,columnspan=4)
                for d in list(enumerate(self.Imat_dict)):
                    mat_listbox.insert("end",d[1])
                
                def _sel_dir():
                    
                     self.destination = Tk.filedialog.askdirectory()
                     
                dir_select =Tk.Button(master=ewin,text='Select Directory',command =_sel_dir)
                dir_select.grid(row=2,column=0)
                name_label = Tk.Label(master=ewin,text='File Lead: ').grid(row=3,column=0)
                name_entry = Tk.Entry(master=ewin,text='UNTITLED')
                name_entry.grid(row=3,column=1)
                
                def _export_now():
                    
                    if len(name_entry.get())>0:
                        file_lead=name_entry.get()
                    else:
                        file_lead = 'UNTITLED'
                    filename = self.destination +'/'+ file_lead
                    
                    metafile = filename + '_params.txt'
                    
                    ind_map = mat_listbox.curselection()[0]
                    map_choice = mat_listbox.get(ind_map)
    #                self.Imat_dict[map_choice].write_meta(metafile)
                    self.Imat_dict[map_choice].save_map(filename)
    #                self.expmnt.write_params(self.meta[map_choice],parfile)
    #                self.expmnt.write_map(self.Imat_dict[map_choice],self.meta['directory'])
                            
                    print('Export Complete')
    
                        
                    
                                    
                exp_button = Tk.Button(master=ewin,text='Export Data',command=_export_now)
                exp_button.grid(row=4,column=0)
                
                def _quit_ewin():
                    ewin.quit()
                    ewin.destroy()
                
                quit_button_exp = Tk.Button(master=ewin,text='Quit',command=_quit_ewin)
                quit_button_exp.grid(row=4,column=1)
                ewin.mainloop()
                
                
                
                
            
                
            export_button = Tk.Button(master=self.root,text="Export",command=_exp_win)
            export_button.grid(row=5,column=6)
            
            quit_button = Tk.Button(master=self.root,text="Quit",command=_quit_win)
            quit_button.grid(row=5,column=8)
    
    
    
            Tk.mainloop()
            
            
    def xyz(X,Y,Z):
         xnow = sum([complex(xi) for xi in X.split('+')])
         ynow = sum([complex(yi) for yi in Y.split('+')])
         znow = sum([complex(zi) for zi in Z.split('+')])
         xr,xi,yr,yi,zr,zi = float(np.real(xnow)),float(np.imag(xnow)),float(np.real(ynow)),float(np.imag(ynow)),float(np.real(znow)),float(np.imag(znow))
         return np.array([xr+1.0j*xi,yr+1.0j*yi,zr+1.0j*zi])/np.linalg.norm(np.array([xr+1.0j*xi,yr+1.0j*yi,zr+1.0j*zi]))

       
